@@ -9,10 +9,14 @@ and diagnostics.
 from dataclasses import dataclass
 from enum import IntEnum, auto
 
+from ..core import TokenKind, IntegerKind
 from ..diagnostics import DID, TokenRange
-from .basic import TokenKind, IntegerKind
-from .macros import BuiltinKind, lex_token_from_builtin_spelling
+from ..parsing import ParserState
+
 from .literals import LiteralInterpreter
+from .macros import BuiltinKind, lex_token_from_builtin_spelling
+
+__all__ = ['ExprParser']
 
 
 @dataclass(slots=True)
@@ -86,7 +90,6 @@ class ExprParser:
         lookahead token.
         '''
         # As per the grammar, comma expressions are not acceptable at the top level.
-        from ..parsing import ParserState
         state = ParserState.from_pp(self.pp)
         return self.parse_conditional_expr(state, True), state.token
 
@@ -104,7 +107,7 @@ class ExprParser:
         '''
         # Start by parsing a unary expression.  Then, if the next binary operator in the
         # token sequence is of precedence less than or equal to reduce_precedence, then
-        # perform (in LR parsing terminology) a reduction operation, oterhwise perform a
+        # perform (in LR parsing terminology) a reduction operation, otherwise perform a
         # shift operation.
         lhs = self.parse_unary_expr(state, is_evaluated)
         while True:
@@ -115,7 +118,7 @@ class ExprParser:
                 return lhs
 
             if precedence == BOP.conditional:
-                lhs = self.parse_conditional_branches(state, token, bool(lhs.value), is_evaluated)
+                lhs = self.parse_conditional_branches(state, token, lhs, is_evaluated)
             else:
                 rhs_is_evaluated = is_evaluated
                 if token.kind == TokenKind.LOGICAL_AND:
@@ -123,21 +126,25 @@ class ExprParser:
                 elif token.kind == TokenKind.LOGICAL_OR:
                     rhs_is_evaluated = rhs_is_evaluated and not bool(lhs.value)
                 rhs = self.parse_binary_expr(state, precedence, rhs_is_evaluated)
-                if not (lhs.is_erroneous or rhs.is_erroneous):
+                if lhs.is_erroneous or rhs.is_erroneous:
+                    lhs.is_erroneous = True
+                else:
                     evaluator(self, lhs, rhs, token, is_evaluated)
                 lhs.loc.end = rhs.loc.end
 
-    def parse_conditional_branches(self, state, token, condition_truth, is_evaluated):
-        '''Parse and evaluate the branches of a conditional operator.'''
+    def parse_conditional_branches(self, state, token, condition, is_evaluated):
+        '''Parse and evaluate the branches of the conditional operator.'''
+        condition_truth = bool(condition.value)
         state.enter_context(token.kind, token.loc)
         lhs = self.parse_expr(state, condition_truth and is_evaluated)
         colon = state.leave_context()
         if colon.kind != TokenKind.COLON:
             return lhs
         rhs = self.parse_conditional_expr(state, not condition_truth and is_evaluated)
-        if not (lhs.is_erroneous or rhs.is_erroneous):
-            self.integer_promotions(lhs, rhs, colon, is_evaluated)
+        if not condition.is_erroneous:
+            self.usual_arithmetic_conversions(lhs, rhs, colon, is_evaluated)
         result = lhs if condition_truth else rhs
+        result.is_erroneous = condition.is_erroneous or lhs.is_erroneous or rhs.is_erroneous
         result.loc.start = lhs.loc.start
         result.loc.end = rhs.loc.end
         return result
@@ -163,7 +170,7 @@ class ExprParser:
         # Unary ops
         if kind in unary_ops:
             rhs = self.parse_unary_expr(state, is_evaluated)
-            if is_evaluated and not rhs.is_erroneous:
+            if not rhs.is_erroneous and is_evaluated:
                 self.evaluate_unary_op(rhs, token)
             rhs.loc.start = token.loc
             return rhs
@@ -174,8 +181,11 @@ class ExprParser:
 
         if kind == TokenKind.STRING_LITERAL:
             self.diag(DID.string_invalid_in_pp_expression, token.loc)
+        elif kind == TokenKind.UNTERMINATED:
+            pass  # Already diagnosed
         else:
             self.diag(DID.expected_expression, token.loc)
+            state.recover(token)
 
         return ExprValue(0, False, True, TokenRange(token.loc, token.loc))
 
@@ -210,20 +220,26 @@ class ExprParser:
                          TokenRange(defined.loc, token.loc))
 
     def parse_has_feature_expr(self, state, macro_token, is_evaluated):
+        '''Parse __has_include() __has_cpp_attribute() operators.'''
         token = state.get_token()
         if token.kind == TokenKind.PAREN_OPEN:
             state.enter_context(token.kind, token.loc)
+            # Get the spelling that is the result of the __has() operator and lex it as a token.
             spelling = self.parse_body_method(macro_token.extra.macro)(state, is_evaluated)
             token = state.leave_context()
             if spelling is not None:
-                lex_token_from_builtin_spelling(self.pp, macro_token, spelling)
-                assert macro_token.kind == TokenKind.NUMBER
-                return self.evaluate_literal(macro_token)
+                token = lex_token_from_builtin_spelling(self.pp, macro_token, spelling,
+                                                        TokenRange(macro_token.loc, token.loc))
+                assert token.kind == TokenKind.NUMBER
+                return self.evaluate_literal(token)
         else:
             self.diag(DID.expected_open_paren, token.loc)
+            state.recover(token)
+        # An erroneous result
         return ExprValue(0, False, True, TokenRange(macro_token.loc, token.loc))
 
     def parse_body_method(self, kind):
+        '''Return the method that is responsible for parsing the part between parentheses.'''
         if kind is BuiltinKind.has_include:
             return self.parse_has_include_body
         elif kind is BuiltinKind.has_cpp_attribute:
@@ -232,6 +248,9 @@ class ExprParser:
             assert False
 
     def expect_identifier(self, state):
+        '''Read an identifier and return is IdentifierInfo.  If it is not an identifier, diagnose
+        that and return None.
+        '''
         token = state.get_token()
         if token.kind == TokenKind.IDENTIFIER:
             return token.extra
@@ -240,6 +259,8 @@ class ExprParser:
         return None
 
     def parse_has_attribute_body(self, state, is_evaluated):
+        '''Parse the parenthesized part of __has_cpp_attribute().  Return the spelling of the
+        result, or None on error.'''
         # C and C++ require an attribute-token, which is either identifer or
         # identifier::identifier.
         scope_name = self.expect_identifier(state)
@@ -260,6 +281,8 @@ class ExprParser:
         return self.pp.has_attribute_spelling(scope_spelling, attrib_spelling)
 
     def parse_has_include_body(self, state, is_evaluated):
+        '''Parse the parenthesized part of __has_include().  Return the spelling of the result,
+        or None on error.'''
         header_token = self.pp.create_header_name(in__has_include=True)
         if header_token is None:
             return None
@@ -283,13 +306,13 @@ class ExprParser:
 
     def evaluate_literal(self, token):
         '''Evaluate a character constant or number.  This needs to be done even in unevaluated
-        contexts.'''
-        value, is_unsigned, is_erroneous = 0, False, False
+        contexts because the signedness of the result matters.'''
         literal = self.literal_interpreter.interpret(token)
         if literal.kind == IntegerKind.error:
-            is_erroneous = True
+            value, is_unsigned, is_erroneous = 0, False, True
         else:
             value, is_unsigned = literal.value, self.pp.target.is_unsigned(literal.kind)
+            is_erroneous = False
         return ExprValue(value, is_unsigned, is_erroneous, TokenRange(token.loc, token.loc))
 
     def evaluate_identifier_expr(self, token, is_evaluated):
@@ -307,24 +330,24 @@ class ExprParser:
         elif token.extra == self.true:
             value = 1
         elif is_evaluated:
-            self.diag(DID.identifier_in_pp_expr, token.loc, [self.pp.token_spelling(token)])
+            self.diag(DID.undefined_identifier, token.loc, [self.pp.token_spelling(token)])
         return ExprValue(value, False, is_erroneous, TokenRange(token.loc, token.loc))
 
     def evaluate_unary_op(self, rhs, op):
         '''Evaluate a unary expression.'''
         kind = op.kind
-        if kind == TokenKind.PLUS:
-            pass
-        elif kind == TokenKind.MINUS:
+        if kind == TokenKind.MINUS:
             if rhs.set(-rhs.get(self.mask), self.mask):
                 self.overflow(rhs, op, [rhs.loc])
         elif kind == TokenKind.LOGICAL_NOT:
             rhs.set_boolean(not rhs.value)
         elif kind == TokenKind.TILDE:
             rhs.value = self.mask - rhs.value
+        else:
+            assert kind == TokenKind.PLUS
 
-    def integer_promotions(self, lhs, rhs, op, is_evaluated):
-        '''Perform the usual arithmetic conversions on lhs and rhs.'''
+    def usual_arithmetic_conversions(self, lhs, rhs, op, is_evaluated):
+        '''Perform the usual arithmetic conversions in-place on lhs and rhs.'''
         if lhs.is_unsigned != rhs.is_unsigned:
             # Find the side to convert to unsigned
             side = rhs if lhs.is_unsigned else lhs
@@ -337,7 +360,7 @@ class ExprParser:
 
     def evaluate_arithmetic(self, lhs, rhs, op, is_evaluated):
         '''Evaluate several kinds of binary expression.'''
-        self.integer_promotions(lhs, rhs, op, is_evaluated)
+        self.usual_arithmetic_conversions(lhs, rhs, op, is_evaluated)
         if not is_evaluated:
             return
         kind = op.kind
@@ -384,8 +407,9 @@ class ExprParser:
                 assert not lhs.set(result, self.mask)
 
     def evaluate_arithmetic_direct(self, lhs, rhs, op, is_evaluated):
-        '''These operate directly on the value with no need for get / set operations.'''
-        self.integer_promotions(lhs, rhs, op, is_evaluated)
+        '''These operate directly on the value with no need for interpretation with get / set
+        operations.'''
+        self.usual_arithmetic_conversions(lhs, rhs, op, is_evaluated)
         if not is_evaluated:
             return
         kind = op.kind
@@ -408,14 +432,15 @@ class ExprParser:
         lhs_value, rhs_value = lhs.get(self.mask), rhs.get(self.mask)
         # Check negative or too large
         if rhs_value < 0:
-            # Undefined behaviour - error
+            # Undefined behaviour - hard error
             self.diag(DID.shift_count_negative, op.loc, [rhs.loc])
             lhs.is_erroneous = True
-        elif rhs_value >= self.width:
-            # Undefined behaviour - error
+            return
+        if rhs_value >= self.width:
+            # Undefined behaviour - we cap the shift
             self.diag(DID.shift_count_too_large, op.loc, [rhs.loc])
-            lhs.is_erroneous = True
-        elif op.kind == TokenKind.LSHIFT:
+            rhs_value = self.width
+        if op.kind == TokenKind.LSHIFT:
             # A masked logical bit-shift where the resulting bit-pattern is then
             # interpreted (in C++23).  In C, and earlier C++, the behaviour is more
             # subtle.
@@ -425,7 +450,7 @@ class ExprParser:
                 # represented in the result's type.  We take the C++ value, but warn in
                 # cases where it is undefined in C.
                 if lhs_value < 0:
-                    self.diag(DID.left_shift_of_negative_value, op.loc, [lhs.loc, 0])
+                    self.diag(DID.shift_of_negative_value, op.loc, [0, lhs.loc])
                 elif value > (self.mask >> 1):
                     self.diag(DID.left_shift_overflows, op.loc, [lhs.loc, rhs.loc])
             lhs.value = value & self.mask
@@ -437,12 +462,12 @@ class ExprParser:
             if lhs_value < 0:
                 # Shift the complement, and complement back.
                 lhs.value = self.mask - ((self.mask - lhs.value) >> rhs_value)
-                self.diag(DID.right_shift_of_negative_value, op.loc, [lhs.loc])
+                self.diag(DID.shift_of_negative_value, op.loc, [1, lhs.loc])
             else:
                 lhs.value >>= rhs_value
 
     def evaluate_logical(self, lhs, rhs, op, is_evaluated):
-        '''Evaluate short-circuiting logical expressions (&& and ||).'''
+        '''Evaluate the logical operators && and ||.'''
         if op.kind == TokenKind.LOGICAL_AND:
             lhs.set_boolean(bool(lhs.value) and bool(rhs.value))
         else:
@@ -451,7 +476,13 @@ class ExprParser:
 
     def evaluate_comma(self, lhs, rhs, op, is_evaluated):
         '''Evaluate a comma expression.'''
-        # Fine from C++11.  Reject in C89, in C99 valid if unevaluated.  C++90?
+        # C++11 started permitting commas in integer constant expressions.  Previously
+        # they were not permitted, as in C.  It seems this was not intended to apply to
+        # the preprocessor
+        # (https://www.open-std.org/jtc1/sc22/wg21/docs/cwg_active.html#1436).  We treat
+        # this as a defect in the standard and diagnose unconditionally but treat the
+        # value appropriately.
+        self.diag(DID.comma_in_pp_expression, op.loc)
         lhs.value = rhs.value
         lhs.is_unsigned = rhs.is_unsigned
         lhs.is_erroneous = rhs.is_erroneous
