@@ -3,64 +3,80 @@
 # All rights reserved.
 #
 
+from dataclasses import dataclass
+from enum import IntEnum, auto
+
+from ..core import Token, TokenKind, TokenFlags, SpecialKind
+from ..diagnostics import BufferRange, Diagnostic, DID
 from ..unicode import (
     name_to_cp, utf8_cp, REPLACEMENT_CHAR, is_NFC, is_valid_codepoint,
     is_control_character, codepoint_to_hex, is_XID_Start, is_XID_Continue,
     is_surrogate
 )
 
-from ..diagnostics import BufferRange, DID
-
-from .basic import (
-    Token, TokenKind, TokenFlags, TokenSource, IdentifierInfo, SpecialKind, HEX_DIGIT_VALUES
-)
-from .literals import printable_form
+from .literals import printable_form, HEX_DIGIT_VALUES
 
 __all__ = ['Lexer']
 
 
-ASCII_DIGITS = {ord(c) for c in '0123456789'}
-ASCII_IDENT_START = set(ord(c) for c in 'ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')
+ASCII_DIGITS = set(b'0123456789')
+ASCII_IDENT_START = set(b'ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz')
 ASCII_IDENT_CONTINUE = set.union(ASCII_IDENT_START, ASCII_DIGITS)
-SIGNS = {ord(c) for c in '+-'}
-EPep = {ord(c) for c in 'EPep'}
-NL_WS = {ord('\r'), ord('\n')}
-NON_NL_WS = {ord(c) for c in ' \t\v\f'}
-
-BASIC_CHARSET = set.union(ASCII_IDENT_CONTINUE,
-                          (ord(c) for c in '\t\v\f \n\r!"#$%&\'()*+,-./:;<=>?@[\\]^`{|}~'))
-DCHARS = set.difference(BASIC_CHARSET, (ord(c) for c in ' ()\\\t\v\f\r\n'))
+SIGNS = set(b'+-')
+EPep = set(b'EPep')
+NL_WS = set(b'\r\n')
+NON_NL_WS = set(b' \t\v\f')
+DELIMITER_INVALID = set(b' ()\\\t\v\f\r\n')
 
 
-class Lexer(TokenSource):
+class CharacterKind(IntEnum):
+    '''Categorizes a source file character.'''
+    # A valid character that can start or continue an identifier
+    valid_start = auto()
+    # A valid character that can continue and identifier
+    valid_continue = auto()
+    # A valid character that cannot appear in identifiers
+    valid_other = auto()
+    # A lexically well-formed UCN sequence that is invalid
+    invalid_ucn = auto()
+    # Invalid UTF-8 - bad encoding, surrogate or over-long
+    invalid_utf8 = auto()
 
-    # The preprocessor is necessary to look up e.g. if an identifier is special, or
-    # an alternative token, or language options.
-    def __init__(self, pp, buff, start_loc):
+
+@dataclass(slots=True)
+class Character:
+    '''A source file character.'''
+    # The valid kinds are spelt as their UTF-8 encoding.  invalid UCNs are spelt as their
+    # UCN spelling.  Invalid UTF-8 is spelled as the UTF-8 encoding of the replacement
+    # character.
+    kind: CharacterKind
+    value: int                # -1 or a valid Unicode codepoint
+    diagnostic: object        # A diagnostic or None.  Only set for invalid_ucn or valid_other.
+
+
+class Lexer:
+    '''A lexer.  It tokenizes a buffer terminated with a NUL.'''
+
+    def __init__(self, pp, buff, start_loc, pp_state):
         assert isinstance(buff, (bytes, bytearray, memoryview))
         assert buff and buff[-1] == 0
         self.pp = pp
         self.buff = buff
         self.start_loc = start_loc
+        self.pp_state = pp_state
         self.cursor = 0
+        self.is_start_of_line = bool(pp_state)
+        self.in_header_name = False
         self.clean = True
 
     def cursor_loc(self):
         return self.cursor + self.start_loc
 
-    def skip_bom(self):
-        bom = '\ufeff'.encode()  # b'\xef\xbb\xbf'
-        return len(bom) if self.buff.startswith(bom) else 0
-
     @staticmethod
     def initialize():
         on_char = {}
-        for n in range(128):
-            on_char[n] = Lexer.on_other
-        for n in range(128, 256):
+        for n in range(256):
             on_char[n] = Lexer.on_identifier
-        for c in ASCII_IDENT_START:
-            on_char[c] = Lexer.on_identifier
         for c in ASCII_DIGITS:
             on_char[c] = Lexer.on_number
         on_char.update({
@@ -103,7 +119,7 @@ class Lexer(TokenSource):
         Lexer.on_char = on_char
 
     def diag(self, did, loc, args=None):
-        self.pp.diag(did, loc + self.start_loc, args)
+        self.pp.emit(self.diagnostic(did, loc, args))
 
     def diag_range(self, did, start, end, args=None):
         '''Diagnose a range of characters.  start and end are byte offsets, and start must not be
@@ -111,7 +127,13 @@ class Lexer(TokenSource):
         the single character at that position is diagnosed.  If end is in the middle of a
         multi-byte character, the diagnosis extends to the end of that character.
         '''
-        self.pp.diag(did, BufferRange(start + self.start_loc, end + self.start_loc), args)
+        self.pp.emit(self.diagnostic_range(did, start, end, args))
+
+    def diagnostic(self, did, loc, args=None):
+        return Diagnostic(did, loc + self.start_loc, args)
+
+    def diagnostic_range(self, did, start, end, args=None):
+        return Diagnostic(did, BufferRange(start + self.start_loc, end + self.start_loc), args)
 
     def read_logical_byte(self, cursor):
         # Return the next byte skipping escaped newlines
@@ -182,36 +204,32 @@ class Lexer(TokenSource):
             c, cursor = self.read_char(cursor - 1)
         return c, cursor
 
-    def get_token(self, token):
-        cursor = self.cursor
-        if cursor == 0:
-            token.flags = TokenFlags.BOL
-            cursor = self.skip_bom()
-        else:
-            token.flags = 0
-        token.extra = None
+    def get_token(self):
+        # Internally to the lexer, all locations are ofsets in the buffer.  They are
+        # adjusted by self.start_loc when interfacing externally - when exiting this
+        # function and by Lexer.diag().
+        token = Token(-1, 0, 0, None)
         buff = self.buff
+        cursor = self.cursor
 
         while True:
-            # Internally to the Lexer, all locations are ofsets in the buffer.  They are
-            # adjusted by self.start_loc when interfacing externally - when exiting
-            # get_token() and by Lexer.diag.
             self.clean = True
-            token.loc = cursor
+            token.loc = cursor  # token.loc is used by some on_char handlers
             c = buff[cursor]
             kind, cursor = self.on_char[c](self, token, cursor + 1)
             if kind != TokenKind.WS:
                 break
 
+        self.is_start_of_line = kind == TokenKind.kw_export_keyword
         token.loc += self.start_loc
         token.kind = kind
         self.cursor = cursor
+        return token
 
     def get_token_quietly(self):
         '''Lex a token, without issuing diagnostics, and return it.'''
         self.pp.ignore_diagnostics += 1
-        token = Token.create()
-        self.get_token(token)
+        token = self.get_token()
         self.pp.ignore_diagnostics -= 1
         return token
 
@@ -243,28 +261,19 @@ class Lexer(TokenSource):
             token.flags |= TokenFlags.WS
         else:
             token.flags &= ~TokenFlags.WS
-        token.flags |= TokenFlags.BOL
+        self.is_start_of_line = True
         return TokenKind.WS, cursor
 
     def on_nul(self, token, cursor):
         if cursor == len(self.buff):
             return TokenKind.EOF, cursor - 1
-        return self.on_other(token, cursor)
-
-    def on_other(self, token, cursor):
-        return TokenKind.OTHER, cursor
+        return self.on_identifier(token, cursor)
 
     def on_backslash(self, token, cursor):
-        is_nl, cursor = self.skip_escaped_newline(cursor)
+        is_nl, ncursor = self.skip_escaped_newline(cursor)
         if is_nl:
-            return TokenKind.WS, cursor
-        kind, ident, ncursor = self.maybe_identifier(token, cursor - 1)
-        if kind is not None:
-            # No alternative tokens begin with a UCN (yet!)
-            assert kind == TokenKind.IDENTIFIER
-            token.extra = ident
-            return kind, ncursor
-        return TokenKind.OTHER, cursor
+            return TokenKind.WS, ncursor
+        return self.on_identifier(token, cursor)
 
     def on_brace_open(self, token, cursor):
         return TokenKind.BRACE_OPEN, cursor
@@ -300,8 +309,11 @@ class Lexer(TokenSource):
         # Handle "# ##"
         c, ncursor = self.read_logical_byte(cursor)
         if c == 35:  # '#'
-            return TokenKind.CONCAT, ncursor
-        return TokenKind.HASH, cursor
+            return TokenKind.HASH_HASH, ncursor
+        if self.is_start_of_line:
+            return TokenKind.DIRECTIVE_HASH, cursor
+        else:
+            return TokenKind.HASH, cursor
 
     def on_not(self, token, cursor):
         # Handle "! !="
@@ -346,7 +358,7 @@ class Lexer(TokenSource):
             return c == 58 or c == 62  # ':' or '>'
 
         # An angled header?
-        if self.pp.in_header_name:
+        if self.pp.in_header_name or self.in_header_name:
             kind, ncursor = self.on_delimited_literal(token, cursor)
             if kind == TokenKind.HEADER_NAME:
                 return kind, ncursor
@@ -356,9 +368,10 @@ class Lexer(TokenSource):
 
         c, ncursor = self.read_logical_byte(cursor)
         if c == 61:  # '='
-            c, cursor = self.read_logical_byte(ncursor)
-            if c == 62:  # '>'
-                return TokenKind.LEG, cursor
+            if self.pp.language.is_cxx():
+                c, cursor = self.read_logical_byte(ncursor)
+                if c == 62:  # '>'
+                    return TokenKind.LEG, cursor
             return TokenKind.LE, ncursor
         if c == 60:  # '<'
             c, cursor = self.read_logical_byte(ncursor)
@@ -391,9 +404,10 @@ class Lexer(TokenSource):
         if c == 61:  # '='
             return TokenKind.MINUS_ASSIGN, ncursor
         if c == 62:  # '>'
-            c, cursor = self.read_logical_byte(ncursor)
-            if c == 42:  # '*'
-                return TokenKind.DEREF_STAR, cursor
+            if self.pp.language.is_cxx():
+                c, cursor = self.read_logical_byte(ncursor)
+                if c == 42:  # '*'
+                    return TokenKind.DEREF_STAR, cursor
             return TokenKind.DEREF, ncursor
         return TokenKind.MINUS, cursor
 
@@ -442,12 +456,16 @@ class Lexer(TokenSource):
             return TokenKind.BRACE_CLOSE, ncursor
         if c != 58:  # ':'
             return TokenKind.MODULUS, cursor
+        # %: or %:%:
         c, cursor = self.read_logical_byte(ncursor)
         if c == 37:  # '%'
             c, cursor = self.read_logical_byte(cursor)
             if c == 58:  # ':'
-                return TokenKind.CONCAT, cursor
-        return TokenKind.HASH, ncursor
+                return TokenKind.HASH_HASH, cursor
+        if self.is_start_of_line:
+            return TokenKind.DIRECTIVE_HASH, ncursor
+        else:
+            return TokenKind.HASH, ncursor
 
     def on_line_comment(self, token, cursor):
         # A line comment.  Delegate handling of EOF and newlines.
@@ -492,20 +510,20 @@ class Lexer(TokenSource):
         c, ncursor = self.read_logical_byte(cursor)
         if c in ASCII_DIGITS:
             return self.on_number(token, cursor)
-        if c == 42:  # '*'
-            return TokenKind.DOT_STAR, ncursor
         if c == 46:  # '.'
             c, ncursor = self.read_logical_byte(ncursor)
             if c == 46:  # '.'
                 return TokenKind.ELLIPSIS, ncursor
+        if c == 42 and self.pp.language.is_cxx():  # '*'
+            return TokenKind.DOT_STAR, ncursor
         return TokenKind.DOT, cursor
 
     # pp-number:
     #     digit
     #    . digit
     #    pp-number identifier-continue
-    #    pp-number’ digit
-    #    pp-number’ nondigit
+    #    pp-number ’ digit
+    #    pp-number ’ nondigit
     #    pp-number e sign
     #    pp-number E sign
     #    pp-number p sign
@@ -519,26 +537,23 @@ class Lexer(TokenSource):
     #    [0-9]
     # non-digit:
     #    [A-Z] [a-z] _
-    #
     def on_number(self, token, cursor):
-        # Handle 0 1 2 3 4 5 6 7 8 9 .digit
+        '''Lex a pp-number.'''
         start = cursor - 1
         c = None
         buff = self.buff
         while True:
+            char_start = cursor
             prevc = c
-            saved_cursor = cursor
             c = buff[cursor]
             cursor += 1
-
             # Fast-track standard ASCII numbers
             if c in ASCII_IDENT_CONTINUE:
                 continue
-
-            # continues_identifier() handles escaped newlines (in which case c is the character
-            # after them), UTF-8 characters, and UCNs
-            is_valid, c, cursor = self.continues_identifier(token, cursor, False, c)
-            if is_valid:
+            character, cursor = self.read_extended_character(cursor - 1)
+            c = character.value
+            if (character.kind is CharacterKind.valid_start
+                    or character.kind is CharacterKind.valid_continue):
                 continue
             if c == 46:  # '.'
                 continue
@@ -550,128 +565,122 @@ class Lexer(TokenSource):
                     continue
 
             if not self.pp.skipping:
-                token.extra = (self.fast_utf8_spelling(start, saved_cursor), None)
-            return TokenKind.NUMBER, saved_cursor
+                token.extra = (self.fast_utf8_spelling(start, char_start), None)
+            return TokenKind.NUMBER, char_start
 
-    def continues_identifier(self, token, cursor, is_ident_start, c):
-        '''Return a triple (is_valid, code_point, cursor).
-
-        If it is not lexically an identifier (e.g. a backslash, or a double quote,
-        is_valid is False and code_point is the character concerned.
-
-        If it is lexically an identifier (e.g. it starts a UCN) then is_valid is true even
-        if lexically incomplete.  code_point is -1 if lexically incomplete, or if the
-        character is semantically invalid.
-        '''
-        char_loc = cursor - 1
-        if c == 92:  # '\\'
-            # Maybe the backslash started an escaped newline
-            c, cursor = self.read_logical_byte(cursor - 1)
-
-        # Handle UCNs
-        if c == 92:  # '\\'
-            # A backslash has been consumed.  A UCN starts / continues the identifier.
-            c, cursor = self.maybe_ucn(cursor, char_loc, is_ident_start, False)
-            if c == -1:
-                return False, c, char_loc
-            self.clean = False
-        elif c < 0x80:
-            valid_chars = ASCII_IDENT_START if is_ident_start else ASCII_IDENT_CONTINUE
-            return c in valid_chars, c, cursor
-        else:
-            # Handle UTF-8 extended identifiers.  Have invalid UTF-8 returned as -1,
-            # and substitute the replacement character.
-            c, cursor = self.read_char(cursor - 1, -1)
-            if c == -1:
-                c = REPLACEMENT_CHAR
-            else:
-                self.validate_codepoint(c, is_ident_start, char_loc, cursor, False)
-
-        return True, c, cursor
-
-    def on_identifier(self, token, cursor):
-        '''Return a (token_kind, cursor) pair.
+    def on_identifier(self, token, cursor, ud_suffix=False):
+        '''Return a (token_kind, cursor) pair.  If lexing a ud-suffix, set it to True.
 
         The token kind is normally TokenKind.IDENTIFIER, but conversion to alternative
         tokens is handled.  In either case token.extra is the IdentifierInfo object.
 
         Encoding-prefixed strings are also handled, in which case token.extra is as for
-        strings.
+        strings.  If it is a non-identifier character, kind is CHARACTER and token.extra
+        is its value.  kind can also be a module keyword.
         '''
-        # Start again at the beginning
-        kind, ident, cursor = self.maybe_identifier(token, token.loc)
-        assert ident is not None
+        cursor -= 1
+        start = cursor
+        buff = self.buff
+        quick_chars = ASCII_IDENT_START
+
+        while True:
+            char_start = cursor
+            c = buff[cursor]
+            cursor += 1
+            # Fast-track standard ASCII identifiers
+            if c not in quick_chars:
+                character, cursor = self.read_extended_character(cursor - 1)
+                if character.kind == CharacterKind.valid_start:
+                    pass
+                elif character.kind == CharacterKind.valid_continue and char_start != start:
+                    pass
+                else:
+                    break
+            quick_chars = ASCII_IDENT_CONTINUE
+
+        if char_start == start:
+            # Emit diagnostics only at the start of a token
+            if character.diagnostic and not self.pp.skipping and not ud_suffix:
+                self.pp.emit(character.diagnostic)
+            token.extra = character.value
+            return TokenKind.CHARACTER, cursor
+
+        cursor = char_start
+        if not self.pp.skipping:
+            spelling = self.fast_utf8_spelling(start, cursor)
+            if not is_NFC(spelling):
+                self.diag(DID.identifier_not_NFC, start, [spelling])
+            ident = self.pp.get_identifier(spelling)
+            token.extra = ident
+            if ident.special:
+                return self.handle_special_identifier(token, start, cursor, ud_suffix)
+
+        return TokenKind.IDENTIFIER, cursor
+
+    def handle_special_identifier(self, token, start, cursor, ud_suffix):
+        '''Handle __VA_ARGS__, alternative tokens, etc.'''
+        ident = token.extra
+        special = ident.special
+        kind = TokenKind.IDENTIFIER
 
         # Is this an encoding prefix to a string or character literal?
-        if kind == TokenKind.IDENTIFIER and ident.special & SpecialKind.ENCODING_PREFIX:
+        if special & SpecialKind.ENCODING_PREFIX and not ud_suffix:
             encoding = ident.encoding()
             c, ncursor = self.read_logical_byte(cursor)
             if ident.spelling[-1] == 82:  # 'R'
                 # There is no such thing as a raw character literal
                 if c == 34:  # '"'
+                    token.extra = None
                     token.flags |= TokenFlags.encoding_bits(encoding)
                     return self.raw_string_literal(token, ncursor)
             elif c == 34 or c == 39:  # '"' and "'"
+                token.extra = None
                 token.flags |= TokenFlags.encoding_bits(encoding)
                 return self.on_delimited_literal(token, ncursor)
-
-        token.extra = ident
+        elif special & SpecialKind.VA_IDENTIFIER:
+            if not self.pp.in_variadic_macro_definition:
+                self.diag(DID.invalid_variadic_identifier_use, start, [ident.spelling])
+        elif special & SpecialKind.ALT_TOKEN:
+            kind = ident.alt_token_kind()
+        elif self.is_start_of_line and special & SpecialKind.MODULE_KEYWORD and not ud_suffix:
+            # The special keyword shall not be an object-like macro
+            if ident.macro and not ident.macro.is_function_like():
+                note = Diagnostic(DID.macro_defined_here, ident.macro.name_loc, [ident.spelling])
+                self.diag(DID.macro_in_module_directive, start, [ident.spelling, note])
+            else:
+                return self.maybe_module_keyword(ident, cursor), cursor
         return kind, cursor
 
-    def maybe_identifier(self, token, cursor):
-        '''Lex what may be an identifier, and return a tuple (kind, ident, cursor).
+    def maybe_module_keyword(self, ident, cursor):
+        '''Called by the lexer on lexing what is perhaps a module keyword.'''
+        mk_kind = ident.module_keyword_kind()
+        if mk_kind == TokenKind.kw_module_keyword:
+            choices = (TokenKind.IDENTIFIER, TokenKind.COLON, TokenKind.SEMICOLON)
+        elif mk_kind == TokenKind.kw_export_keyword:
+            choices = (TokenKind.kw_module_keyword, TokenKind.kw_import_keyword)
+        else:
+            assert mk_kind == TokenKind.kw_import_keyword
+            choices = (TokenKind.HEADER_NAME, TokenKind.LT, TokenKind.IDENTIFIER,
+                       TokenKind.STRING_LITERAL, TokenKind.COLON)
 
-        If it is lexically an identifier, ident is an IdentifierInfo object.  If that
-        identifier becomes an alternative token, kind is that token kind, oterhwise it is
-        TokenKind.IDENTIFIER.
+        assert not self.in_header_name
+        was_in_directive = self.pp.in_directive
+        self.pp.in_directive = True               # To obtain EOF at end-of-line
+        self.in_header_name = mk_kind == TokenKind.kw_import_keyword
+        self.cursor = cursor
+        peeked_kind = self.peek_token_kind()
+        # Ensure a subsequent lex gets the header name
+        self.in_header_name = peeked_kind == TokenKind.HEADER_NAME
+        self.pp.in_directive = was_in_directive
+        return mk_kind if peeked_kind in choices else TokenKind.IDENTIFIER
 
-        If it is not an identifier kind is None and cursor is unchanged from the one
-        passed in.
-        '''
-        start = cursor
-        buff = self.buff
-        is_start = True
-        quick_chars = ASCII_IDENT_START
-
-        while True:
-            c = buff[cursor]
-            ncursor = cursor + 1
-            # Fast-track standard ASCII identifiers
-            if c not in quick_chars:
-                is_valid, _c, ncursor = self.continues_identifier(token, ncursor, is_start, c)
-                if not is_valid:
-                    break
-            cursor = ncursor
-            is_start = False
-            quick_chars = ASCII_IDENT_CONTINUE
-
-        if cursor == start:
-            return None, None, cursor
-
-        kind = TokenKind.IDENTIFIER
-        if self.pp.skipping:
-            return kind, IdentifierInfo.dummy, cursor
-
-        spelling = self.fast_utf8_spelling(start, cursor)
-        if not is_NFC(spelling):
-            self.diag(DID.identifier_not_NFC, start, [spelling])
-
-        # Handle __VA_ARGS__, alternative tokens, etc.
-        ident = self.pp.get_identifier(spelling)
-        if ident.special & SpecialKind.VA_IDENTIFIER:
-            if not self.pp.in_variadic_macro_definition:
-                self.diag(DID.invalid_variadic_identifier_use, start, [spelling])
-        elif ident.special & SpecialKind.ALT_TOKEN:
-            kind = ident.alt_token_kind()
-
-        return kind, ident, cursor
-
-    # Character and (non-raw) string literals
     def on_delimited_literal(self, token, cursor):
-        '''The encoding prefix, if any, and the opening quote are already lexed.'''
+        '''Character literals, non-raw string literals, and header names.'''
+        # The encoding prefix, if any, and the opening quote are already lexed
         buff = self.buff
         delimeter = buff[cursor - 1]
-        in_header = self.pp.in_header_name
+        in_header = self.pp.in_header_name or self.in_header_name
+        self.in_header_name = False
 
         if delimeter == 34:     # '"'
             kind = TokenKind.HEADER_NAME if in_header else TokenKind.STRING_LITERAL
@@ -714,8 +723,8 @@ class Lexer(TokenSource):
             if not in_header and not self.pp.skipping:
                 selector = 0 if kind == TokenKind.CHARACTER_LITERAL else 1
                 self.diag(DID.unterminated_literal, token.loc, [selector])
-            # Unterminated literals become the error token
-            return TokenKind.ERROR, cursor
+            token.extra = None
+            return TokenKind.UNTERMINATED, cursor
 
         # Header names have no ud_suffix
         if in_header:
@@ -736,11 +745,11 @@ class Lexer(TokenSource):
 
         Otherwise return (ident, cursor) where ident is an IdentifierIno object.
         '''
-        kind, ident, ncursor = self.maybe_identifier(token, cursor)
-        # An alternative token must not be treated as a user-defined suffix.
-        if kind != TokenKind.IDENTIFIER:
-            return None, cursor
-        return ident, ncursor
+        if self.pp.language.is_cxx():
+            kind, ncursor = self.on_identifier(token, cursor + 1, ud_suffix=True)
+            if kind == TokenKind.IDENTIFIER:
+                return token.extra, ncursor
+        return None, cursor
 
     # raw-string:
     #     "d-char-sequence[opt] (r-char-sequence[opt]) d-char-sequence[opt]"
@@ -761,11 +770,11 @@ class Lexer(TokenSource):
         '''The encoding prefix, if any, and the opening quote are already lexed.  The token began
         at token_loc.
         '''
-        def lex_delimeter(buff, cursor):
+        def lex_delimeter(buff, cursor, basic_charset):
             '''Return the byte terminating the delimeter, and the cursor position beyond it.'''
             while True:
                 c = buff[cursor]
-                if c not in DCHARS:
+                if c not in basic_charset or c in DELIMITER_INVALID:
                     return c, cursor
                 cursor += 1
 
@@ -774,41 +783,39 @@ class Lexer(TokenSource):
         buff = self.buff
         diagnose = not self.pp.skipping
         delim_start = cursor
-        c, cursor = lex_delimeter(buff, cursor)
+        c, cursor = lex_delimeter(buff, cursor, self.pp.basic_charset)
         delimeter = buff[delim_start: cursor]
         # This seems arbitrary and pointless requirement...
         if len(delimeter) > 16 and diagnose:
             self.diag_range(DID.delimeter_too_long, delim_start, cursor)
 
-        if c == 40:  # '(':
-            # Lex the raw part
-            delimeter += bytes([34])  # '"'
-            cursor += 1
-            while True:
-                c, cursor = self.read_char(cursor)
-                if c == 0 and cursor == len(self.buff):
-                    if diagnose:
-                        self.diag(DID.unterminated_literal, token.loc, [2])
-                    # Unterminated literals become the error token
-                    return TokenKind.ERROR, cursor - 1
-                # ')'
-                if c == 41 and buff[cursor: cursor + len(delimeter)] == delimeter:
-                    cursor += len(delimeter)
-                    break
-        elif diagnose:
+        if c != 40:  # '(':
             is_eof = c == 0 and cursor + 1 == len(self.buff)
-            bad_loc = cursor
             if is_eof:
                 c = 10
             else:
-                c, cursor = self.read_char(cursor)
-            self.diag_range(DID.delimeter_invalid_character, bad_loc, bad_loc,
-                            [printable_form(c)])
+                c, _ = self.read_char(cursor)
+            if diagnose:
+                self.diag_range(DID.delimeter_invalid_character, cursor, cursor,
+                                [printable_form(c)])
             # Recover by skipping to end-of-line or EOF.  Note this will find ill-formed UTF-8.
             while c != 10 and c != 13 and cursor != len(self.buff):
                 c, cursor = self.read_char(cursor)
-            # Unterminated literals become the error token
-            return TokenKind.ERROR, cursor
+            return TokenKind.UNTERMINATED, cursor
+
+        # Lex the raw part
+        delimeter += bytes([34])  # '"'
+        cursor += 1
+        while True:
+            c, cursor = self.read_char(cursor)
+            if c == 0 and cursor == len(self.buff):
+                if diagnose:
+                    self.diag(DID.unterminated_literal, token.loc, [2])
+                return TokenKind.UNTERMINATED, cursor - 1
+            # ')'
+            if c == 41 and buff[cursor: cursor + len(delimeter)] == delimeter:
+                cursor += len(delimeter)
+                break
 
         ud_suffix, cursor = self.maybe_user_defined_suffix(token, cursor)
         # No point calling fast_utf8_spelling
@@ -817,6 +824,28 @@ class Lexer(TokenSource):
 
         return TokenKind.STRING_LITERAL, cursor
 
+    #
+    # Extended characters and UCNs
+    #
+
+    def read_extended_character(self, cursor):
+        '''Return a pair (character, cursor).'''
+        c, cursor = self.read_logical_byte(cursor)
+        if c < 0x80:
+            if c in ASCII_IDENT_START:
+                return Character(CharacterKind.valid_start, c, None), cursor
+            if c in ASCII_IDENT_CONTINUE:
+                return Character(CharacterKind.valid_continue, c, None), cursor
+            if c == 92:  # '\\'
+                self.clean = False
+                return self.maybe_ucn(cursor)
+            return Character(CharacterKind.valid_other, c, None), cursor
+        # Handle UTF-8 extended identifiers
+        c, cursor = self.read_char(cursor - 1, -1)
+        if c == -1:
+            return Character(CharacterKind.invalid_utf8, -1, None), cursor
+        return self.extended_character(c), cursor
+
     # universal-character-name:
     #     \u hex-quad
     #     \U hex-quad hex-quad
@@ -829,17 +858,14 @@ class Lexer(TokenSource):
     # simple-hexadecimal-digit-sequence:
     #    hexadecimal-digit
     #    simple-hexadecimal-digit-sequence hexadecimal-digit
-    #
-    def hex_ucn(self, cursor, is_U):
-        '''Try to lex the portion of a hexadecimal univeral character after the backslash-u or
-        backslash-U prefix has been consumed.
-
-        Return a pair (cp, cursor).  If lexically invalid (insufficient digits, no closing
-        brace), codepoint, cp is -1.  Otherwise cp is the hexadecimal value.
+    def hex_ucn(self, cursor, is_U, ucn_start):
+        '''Lex a hexadecimal UCN; backslash-u or U has been consumed.  Return a (character,
+        cursor) pair.
         '''
         count = 0
         cp = 0
         limit = 8 if is_U else 4
+        u_cursor = cursor - 1
 
         while True:
             c, cursor = self.read_logical_byte(cursor)
@@ -863,9 +889,11 @@ class Lexer(TokenSource):
             else:
                 cursor -= 1
 
-            return -1, cursor
+            # This was an incomplete UCN.  Continue future lexing from the 'u' or 'U'.
+            diagnostic = self.diagnostic_range(DID.incomplete_UCN_as_tokens, ucn_start, cursor)
+            return Character(CharacterKind.valid_other, 92, diagnostic), u_cursor
 
-        return cp, cursor
+        return self.ucn_character(cp, ucn_start, cursor)
 
     # named-universal-character:
     #    \N{ n-char-sequence }
@@ -874,73 +902,74 @@ class Lexer(TokenSource):
     # n-char-sequence:
     #    n-char
     #    n-char-sequence n-char
-    def named_character(self, cursor, quiet):
-        '''Try to lex the portion of a named univeral character after the backslash-N prefix
-        has been consumed.
-
-        Return a triple (lexically_valid, value, cursor).  If lexically invalid, value is
-        -1 and cursor is the lexically invalid character.  If lexically invalid but the
-        named character is unknown (which is diagnosed) then value is -1.
+    def named_character(self, cursor, ucn_start):
+        '''Lex a named univeral character; backslash-N has been consumed.  Return a (character,
+        cursor) pair.
         '''
-        c, cursor = self.read_logical_byte(cursor)
-        if c != 123:  # '{'
-            cursor -= 1
-        else:
-            name = ''
-            name_loc = cursor
-            while True:
-                c, cursor = self.read_logical_byte(cursor)
-                if c == 125:  # '}'
+        N_cursor = cursor - 1
+        name = ''
+        name_loc = None
+        while True:
+            char_start = cursor
+            c, cursor = self.read_logical_char(cursor)
+            if name_loc is None:
+                if c != 123:  # '{'
                     break
-                if c == 0 and cursor == len(self.buff):
-                    cursor -= 1
-                    break
-                if c == 10 or c == 13:  # '\n' '\r'
-                    # Don't swallow the newline sequence
-                    cursor -= 1
-                    break
-                name += chr(c)
+                name_loc = cursor
+                continue
+            # '}' '\n' '\r'
+            if c == 125 or c == 10 or c == 13 or (c == 0 and cursor == len(self.buff)):
+                break
+            name += chr(c)
 
-            if c == 125 and name:  # '}'
-                cp = name_to_cp(name)
-                if cp == -1 and not self.pp.skipping and not quiet:
-                    self.diag_range(DID.unrecognized_universal_character_name,
-                                    name_loc, cursor - 1, [name])
-                return True, cp, cursor
+        if c == 125 and name:  # '}'
+            cp = name_to_cp(name)
+            if cp != -1:
+                return self.ucn_character(cp, ucn_start, cursor)
+            diagnostic = self.diagnostic_range(DID.unrecognized_universal_character_name,
+                                               name_loc, cursor - 1, [name])
+            return Character(CharacterKind.invalid_ucn, -1, diagnostic), cursor
 
-        return False, -1, cursor
-
-    def maybe_ucn(self, cursor, escape_loc, is_ident_start, quiet):
-        '''On entry, a backslash has been consumed.  Return a triple (cp, cursor).
-
-        If lexically the backslash does not form a UCN, then cp is -1.  If lexically it
-        does form a UCN, then the codepoint is checked for validity in context.  If the
-        codepoint is valid it is returned, otherise REPLACEMENT_CHAR is returned.
-        '''
-        saved_cursor = cursor
+        # This was an incomplete UCN.  Continue future lexing from the 'N'.
+        diagnostic = self.diagnostic_range(DID.incomplete_UCN_as_tokens, ucn_start, char_start)
+        return Character(CharacterKind.valid_other, 92, diagnostic), N_cursor
 
+    def maybe_ucn(self, cursor):
+        '''Called after a logical backslash.  Work out what the character really is.'''
+        ucn_start = cursor - 1
         c, cursor = self.read_logical_byte(cursor)
         if c == 85 or c == 117:  # 'U' 'u'
-            cp, cursor = self.hex_ucn(cursor, c == 85)
-            is_ucn = cp != -1
-        elif c == 78:  # 'N'
-            is_ucn, cp, cursor = self.named_character(cursor, quiet)
-        else:
-            # It didn't begin anything resembling a UCN sequence
-            return -1, saved_cursor
+            return self.hex_ucn(cursor, c == 85, ucn_start)
+        if c == 78:  # 'N'
+            return self.named_character(cursor, ucn_start)
+        # Not a UCN, just the backslash.  Later lexing can start at the logical byte
+        return Character(CharacterKind.valid_other, 92, None), cursor - 1
 
-        if is_ucn:
-            # -1 is an unknown named character.
-            if cp != -1:
-                cp = self.validate_codepoint(cp, is_ident_start, escape_loc, cursor, quiet)
-            if cp == -1:
-                cp = REPLACEMENT_CHAR
+    def ucn_character(self, c, ucn_start, cursor):
+        '''Validate the value C of a UCN over the range ucn_start to cursor.  Return a (character,
+        cursor) pair.
+        '''
+        if not is_valid_codepoint(c):
+            did = DID.codepoint_invalid
+        elif is_surrogate(c):
+            did = DID.codepoint_surrogate
+        elif is_control_character(c):
+            did = DID.codepoint_control_character
+        elif c in self.pp.basic_charset:
+            did = DID.codepoint_basic_character_set
         else:
-            assert cp == -1
-            if is_ident_start and not quiet and not self.pp.skipping:
-                self.diag_range(DID.incomplete_UCN_as_tokens, escape_loc, cursor)
+            return self.extended_character(c), cursor
+        diagnostic = self.diagnostic_range(did, ucn_start, cursor, [codepoint_to_hex(c)])
+        return Character(CharacterKind.invalid_ucn, -1, diagnostic), cursor
 
-        return cp, cursor
+    def extended_character(self, c):
+        '''Return a Character instance representing an extended character of value 'c'.'''
+        if is_XID_Start(c):
+            return Character(CharacterKind.valid_start, c, None)
+        elif is_XID_Continue(c):
+            return Character(CharacterKind.valid_continue, c, None)
+        else:
+            return Character(CharacterKind.valid_other, c, None)
 
     def token_spelling_at_cursor(self):
         prior_cursor = self.cursor
@@ -1025,13 +1054,14 @@ class Lexer(TokenSource):
 
             if cp < 0x80:
                 if cp == 92 and delim is None and rs_limit is None:
-                    cp, cursor = self.maybe_ucn(cursor, cursor - 1, count == 0, True)
-                    if cp == -1:
-                        cp = 92
-                if cp < 0x80:
+                    character, cursor = self.maybe_ucn(cursor)
+                    if character.kind == CharacterKind.invalid_ucn:
+                        encoding = self.clean_spelling(char_start, cursor)
+                    else:
+                        encoding = chr(character.value).encode()
+                else:
                     result.append(cp)
                     continue
-                encoding = chr(cp).encode()
             else:
                 cursor -= 1
                 cp, size = utf8_cp(self.buff, cursor)
@@ -1052,51 +1082,12 @@ class Lexer(TokenSource):
 
         return bytes(result)
 
-    def validate_codepoint(self, cp, is_ident_start, start, end, quiet):
-        assert cp != -1
-        did = did_for_codepoint(cp, is_ident_start)
-        if did:
-            if not quiet and not self.pp.skipping:
-                self.diag_range(did, start, end, [codepoint_to_hex(cp)])
-            cp = REPLACEMENT_CHAR
-        return cp
-
-
-# Requirements on UCNs in the C++23 standard:
-#
-# 1) \u \U UCNs must be a unicode scalar value (in-range and not a surrogate)
-# 2) \N must be its character name in UCD, or "control/correction/alternate" in NameAliases.txt
-# 3) Outside a c-char, s-char or r-char sequence, of a character or string literal, is a control
-#    character or in the basic character set, the program is ill-formed.
-# 4) Identifier characters must have XID_Start and XID_Continue properties
-# 5) Identifier needs to conform to Normalization Form C.
-def did_for_codepoint(cp, is_ident_start):
-    '''Diagnose bad UCN values.  Return True if valid.'''
-    assert cp >= 0
-
-    if not is_valid_codepoint(cp):
-        return DID.codepoint_invalid
-
-    if is_surrogate(cp):
-        return DID.codepoint_surrogate
-
-    # If a universal-character-name outside the c-char-sequence, s-char-sequence, or
-    # r-char-sequence of a character-literal or string-literal (in either case,
-    # including within a user-defined-literal) corresponds to a control character or
-    # to a character in the basic character set, the program is ill-formed.
-    if is_control_character(cp):
-        return DID.codepoint_control_character
-
-    if cp in BASIC_CHARSET:
-        return DID.codepoint_basic_character_set
-
-    if is_ident_start:
-        if not is_XID_Start(cp):
-            return DID.codepoint_cannot_begin_identifier
-    elif not is_XID_Continue(cp):
-        return DID.codepoint_cannot_continue_identifier
-
-    return None
+    def clean_spelling(self, cursor, end):
+        result = bytearray()
+        while cursor < end:
+            c, cursor = self.read_logical_char(cursor)
+            result.extend(chr(c).encode())
+        return result
 
 
 Lexer.initialize()
