@@ -4,23 +4,19 @@
 #
 '''Interpretation of character, string and numeric literals.'''
 
-from copy import copy
 from dataclasses import dataclass
 from enum import IntEnum
 from struct import Struct
 
-from ..core import IntegerKind, RealKind
+from ..core import (
+    IntegerKind, RealKind, TokenKind, TokenFlags, Encoding, IdentifierInfo, Token,
+)
 from ..diagnostics import DID, SpellingRange, Diagnostic
 from ..unicode import (
     utf8_cp, printable_char, is_surrogate, is_valid_codepoint, name_to_cp,
     codepoint_to_hex, SIMPLE_ESCAPES, CONTROL_CHARACTER_LETTERS, Charset,
 )
 
-from .basic import (
-    HEX_DIGIT_VALUES, TokenKind, TokenFlags, Encoding, IdentifierInfo, Token, value_width
-)
-
-
 __all__ = [
     'LiteralInterpreter', 'IntegerLiteral', 'FloatingPointLiteral', 'StringLiteral',
     'printable_form', 'destringize',
@@ -28,10 +24,22 @@ __all__ = [
 
 
 OCTAL_DIGITS = {ord(c) for c in '01234567'}
+DIGIT_VALUES = {ord(c): ord(c) - 48 for c in '0123456789'}
+HEX_DIGIT_VALUES = {
+    ord('a'): 10, ord('b'): 11, ord('c'): 12, ord('d'): 13, ord('e'): 14, ord('f'): 15,
+    ord('A'): 10, ord('B'): 11, ord('C'): 12, ord('D'): 13, ord('E'): 14, ord('F'): 15,
+}
+HEX_DIGIT_VALUES.update(DIGIT_VALUES)
 # Map bases to diagnostic arguments for DID.invalid_digit
 bases = [2, 8, 10, 16]
 
 
+def value_width(value):
+    if value >= 0:
+        return value.bit_length()
+    return (-value - 1).bit_length() + 1
+
+
 @dataclass(slots=True)
 class UserDefinedSuffix:
     '''A user-defined suffix.  The location is provided in case it needs to be diagnosed.'''
@@ -104,7 +112,7 @@ class StringLiteral:
 
     def to_short_text(self):
         ud_suffix_part = f', {self.ud_suffix.ident.spelling.decode()}' if self.ud_suffix else ''
-        return f'StringLiteral({self.encoded.decode()}, {self.char_kind.name}{ud_suffix_part})'
+        return f'StringLiteral({bytes(self.encoded)}, {self.char_kind.name}{ud_suffix_part})'
 
 
 def printable_form(cp):
@@ -211,13 +219,6 @@ class LiteralInterpreter:
 
     def interpret(self, token):
         '''Return a literal.'''
-        if token.kind == TokenKind.STRING_LITERAL:
-            # Reject string literals in preprocessor expressions
-            if self.pp_arithmetic:
-                self.pp.diag(DID.string_invalid_in_pp_expression, token.loc)
-                return IntegerLiteral(IntegerKind.error, 0, None)
-            return self.concatenate_string_literals(token)
-
         if token.kind == TokenKind.NUMBER:
             result = self.interpret_number(token)
         else:
@@ -678,8 +679,12 @@ class LiteralInterpreter:
         else:
             return SpellingRange(token.loc, limit + 1, len(spelling))
 
-    def concatenate_string_literals(self, token):
-        '''Concatenate and interpret string literals beginning with token.'''
+    def concatenate_strings(self, token):
+        '''Concatenate and interpret consecutive string literals beginning with token.
+
+        Return a pair (literal, next_token).  literal is an instance of StringLiteral, and
+        next_token is the first token consumed that was not a string literal.
+        '''
         def diagnose_conflict(token, bad_tokens, selector):
             args = [selector]
             args.extend([Diagnostic(DID.string_concatenation_prior,
@@ -689,14 +694,16 @@ class LiteralInterpreter:
             self.pp.diag(DID.string_concatenation_conflict, diag_loc, args)
 
         assert token.kind == TokenKind.STRING_LITERAL
+        assert not self.pp_arithmetic
 
         # Read adjacent string literal tokens.  This leaves the next non-string-literal
         # token in token.
         is_erroneous = False
         tokens = []
         while token.kind == TokenKind.STRING_LITERAL:
-            tokens.append(copy(token))
-            self.pp.get_token(token)
+            tokens.append(token)
+            token = self.pp.get_token()
+        next_token = token
 
         # Determine a common encoding prefix and user-defined suffix.  All concatenated
         # tokens with an encoding prefix must have the same one, and similarly all with a
@@ -743,7 +750,7 @@ class LiteralInterpreter:
         encoded.extend(encoder('', final=True) + encoder('\0'))
 
         char_kind = IntegerKind.error if is_erroneous else elab_encoding.char_kind
-        return StringLiteral(encoded, char_kind, common_ud_suffix)
+        return StringLiteral(encoded, char_kind, common_ud_suffix), next_token
 
     def encode_string(self, token, encoded, encoding, char_width):
         '''Encode the string literal 'token' into 'encoded', a bytearray.  'encoding' contains
@@ -912,7 +919,7 @@ class LiteralInterpreter:
                 return -1, cursor
 
             if value != -1:
-                state.diag_char(DID.invalid_digit, c_cursor, [1])
+                state.diag_char(DID.invalid_digit, c_cursor, [bases.index(radix)])
                 value = -1
 
     def maybe_ucn(self, state, cursor, cp):
