@@ -10,10 +10,11 @@ from bisect import bisect_left
 from dataclasses import dataclass
 from enum import IntEnum, auto
 
+from .lexer import Lexer
 from ..core import Buffer, PresumedLocation
 from ..diagnostics import (
-    BufferRange, TokenRange, SpellingRange, DiagnosticContext, DID, location_none,
-    RangeCoords, DiagnosticSeverity,
+    BufferRange, TokenRange, SpellingRange, DiagnosticContext, DID, RangeCoords,
+    DiagnosticSeverity,
 )
 
 
@@ -40,20 +41,20 @@ class BufferSpan:
     directive (the actual "include" token) so that an include stack can be produced.
     '''
 
-    def __init__(self, buffer, start, parent_loc, name):
+    def __init__(self, buffer, start, name):
         self._buffer = buffer
         self.start = start
         # End is inclusive, so this permits an end-of-buffer location
         self.end = start + len(buffer.text)
-        self._parent_loc = parent_loc
         self.line_ranges = [LineRange(0, name, 0)]
 
     def buffer(self):
         return self._buffer
 
-    def macro_parent_loc(self, loc):
+    def macro_parent_range(self, loc):
         assert self.start <= loc <= self.end
-        return -1
+        # Buffers are terminal
+        return None
 
     def spelling_loc(self, loc):
         return loc
@@ -79,18 +80,19 @@ class BufferSpan:
 
 
 class ScratchBufferSpan(Buffer):
-    '''A scratch buffer holds the spelling of virtual tokens that are generated by the macro
-    expansion process - namely concatenated tokens and stringized tokens.
+    '''A scratch buffer holds the spelling of virtual tokens that are generated during
+    preprocessing - including concatenated tokens, stringized tokens, _Pragma destringized
+    strings, header names, builtin macro tokens, etc.
 
     The locator creates scratch buffers on demand; there may be several or none in
-    existence.  A scratch buffer cannot change size once craeted, as its span of locations
+    existence.  A scratch buffer cannot change size once created as its span of locations
     is fixed and not unbounded.  The scratch buffer keeps track of the origin of the
     virtual tokens it creates so that the macro stack can be correctly reported.
     '''
 
     def __init__(self, start, end):
         '''Create a scratch buffer with the given size.'''
-        super().__init__(bytearray())
+        super().__init__(bytearray(1))
         assert start <= end
         self.start = start
         self.end = end
@@ -101,20 +103,22 @@ class ScratchBufferSpan(Buffer):
         return self
 
     def has_room(self, size):
-        return self.start + len(self.text) + size + 1 <= self.end
+        return len(self.text) + size + 1 <= self.end - self.start
 
-    def add_spelling(self, spelling, parent_loc, entry_kind):
+    def add_spelling(self, spelling, parent_range, entry_kind):
         text = self.text
-        start = self.start + len(text)
         # We place a newline character at the end of the spelling so it appears on its own
         # line in diagnostics.
+        text.pop()  # Drop the NUL
+        loc = self.start + len(text)
         text.extend(spelling)
         text.append(10)
-        assert start + len(text) <= self.end
-        self.entries.append(ScratchEntry(start, parent_loc, entry_kind))
+        text.append(0)
+        assert len(text) <= self.end - self.start
+        self.entries.append(ScratchEntry(loc, parent_range, entry_kind))
         # Clear any cached line offsets
         self._sparse_line_offsets = None
-        return start
+        return loc
 
     def did_and_substitutions(self, pp, loc):
         entry = self.entry_for_loc(loc)
@@ -123,7 +127,8 @@ class ScratchBufferSpan(Buffer):
         elif entry.kind == ScratchEntryKind.stringize:
             return DID.in_argument_stringizing, []
         elif entry.kind == ScratchEntryKind.builtin:
-            return DID.in_expansion_of_builtin, [pp.token_spelling_at_loc(entry.parent_loc)]
+            spelling = pp.token_spelling_at_loc(entry.parent_range.start)
+            return DID.in_expansion_of_builtin, [spelling]
         elif entry.kind == ScratchEntryKind.header:
             return DID.from_formation_of_header_name, []
         elif entry.kind == ScratchEntryKind.pragma:
@@ -140,8 +145,8 @@ class ScratchBufferSpan(Buffer):
     def spelling_loc(self, loc):
         return loc
 
-    def macro_parent_loc(self, loc):
-        return self.entry_for_loc(loc).parent_loc
+    def macro_parent_range(self, loc):
+        return self.entry_for_loc(loc).parent_range
 
     def presumed_location(self, offset):
         '''Return a PresumedLocation instance.'''
@@ -153,6 +158,7 @@ class ScratchBufferSpan(Buffer):
 
 
 class ScratchEntryKind(IntEnum):
+    '''The reason a scratch entry was created.'''
     concatenate = auto()
     stringize = auto()
     builtin = auto()
@@ -162,16 +168,20 @@ class ScratchEntryKind(IntEnum):
 
 @dataclass(slots=True)
 class ScratchEntry:
+    '''Describes an entry in the scratch buffer(s).'''
     offset: int
-    parent_loc: int
+    parent_range: TokenRange
     kind: ScratchEntryKind
 
 
 @dataclass(slots=True)
 class MacroReplacementSpan:
-
+    '''Represents the tokens of a macro's replacement list before parameter replacement takes
+    place, in other words, as it appeared in the #define directive.  There is one location
+    per token.
+    '''
     macro: object
-    invocation_loc: int
+    invocation_range: TokenRange
     start: int
     end: int
 
@@ -179,38 +189,37 @@ class MacroReplacementSpan:
         token_index = loc - self.start
         return self.macro.replacement_list[token_index].loc
 
-    def macro_parent_loc(self, loc):
-        return self.invocation_loc
-
-    def macro_name(self, pp):
-        '''Return the macro name (as UTF-8).'''
-        return pp.token_spelling_at_loc(self.invocation_loc)
+    def macro_parent_range(self, loc):
+        return self.invocation_range
 
     def did_and_substitutions(self, pp, loc):
-        return DID.in_expansion_of_macro, [self.macro_name(pp)]
+        return DID.in_expansion_of_macro, [self.macro.macro_name(pp)]
 
 
 @dataclass(slots=True)
 class MacroArgumentSpan:
-
+    '''Represents the tokens that replaced a macro parameter in the replacement list.'''
     parameter_loc: int
     start: int
     end: int
     locations: list
 
     def spelling_loc(self, loc):
+        assert self.start <= loc <= self.end
         return self.locations[loc - self.start]
 
-    def macro_parent_loc(self, loc):
-        return self.parameter_loc
+    def macro_parent_range(self, loc):
+        return TokenRange(self.parameter_loc, self.parameter_loc)
 
 
 class Locator:
-    '''Manages and supplies token locations.'''
+    '''Manages and supplies token locations.  A location is an integer but encodes the source
+    of the token and its macro expansion history.
+    '''
 
     FIRST_BUFFER_LOC = 1
     FIRST_MACRO_LOC = 1 << 40
-    prior_file_name = str
+    prior_file_name = str   # Used only as a sentinel
 
     def __init__(self, pp):
         self.pp = pp
@@ -223,9 +232,9 @@ class Locator:
             return self.buffer_spans[-1].end + 1
         return self.FIRST_BUFFER_LOC
 
-    def new_buffer_loc(self, buffer, name, parent_loc):
+    def new_buffer_loc(self, buffer, name):
         start = self.next_buffer_span_start()
-        self.buffer_spans.append(BufferSpan(buffer, start, parent_loc, name))
+        self.buffer_spans.append(BufferSpan(buffer, start, name))
         # Create a scratch buffer after the initial buffer to keep first locations stable
         # if predefines change
         if not self.scratch_range:
@@ -244,13 +253,15 @@ class Locator:
         except IndexError:
             return self.FIRST_MACRO_LOC
 
-    def macro_replacement_span(self, macro, parent_loc):
+    def macro_replacement_span(self, macro, invocation_range):
+        assert isinstance(invocation_range, TokenRange)
         start = self.next_macro_span_start()
         end = start + len(macro.replacement_list) - 1
-        self.macro_spans.append(MacroReplacementSpan(macro, parent_loc, start, end))
+        self.macro_spans.append(MacroReplacementSpan(macro, invocation_range, start, end))
         return start
 
     def macro_argument_span(self, parameter_loc, locations):
+        assert isinstance(parameter_loc, int)
         start = self.next_macro_span_start()
         end = start + len(locations) - 1
         self.macro_spans.append(MacroArgumentSpan(parameter_loc, start, end, locations))
@@ -263,11 +274,12 @@ class Locator:
         self.buffer_spans.append(scratch_range)
         return scratch_range
 
-    def new_scratch_token(self, spelling, parent_loc, entry_kind):
+    def new_scratch_token(self, spelling, parent_range, entry_kind):
+        assert isinstance(parent_range, TokenRange)
         size = len(spelling)
         if not self.scratch_range.has_room(size):
             self.scratch_range = self.create_scratch_range(size)
-        return self.scratch_range.add_spelling(spelling, parent_loc, entry_kind)
+        return self.scratch_range.add_spelling(spelling, parent_range, entry_kind)
 
     def add_line_range(self, start_loc, name, line_number):
         span = self.lookup_span(start_loc)
@@ -297,29 +309,17 @@ class Locator:
                 continue
             return span, loc - span.start
 
-    def buffer_text_and_offset(self, loc):
-        '''Return a buffer's text and offset into it so that the token can be lexed.'''
+    def lexer_at_loc(self, loc):
+        '''Return a new lexer ready to lex the spelling of the token at loc.'''
         span, offset = self.spelling_span_and_offset(loc)
-        return span.buffer().text, offset
+        lexer = Lexer(self.pp, span.buffer().text, loc - offset, None)
+        lexer.cursor = offset
+        return lexer
 
     def derives_from_macro_expansion(self, loc):
         '''Return True if loc is from a macro expansion.'''
         return not isinstance(self.lookup_span(loc), BufferSpan)
 
-    def buffer_span_loc(self, loc):
-        '''Step up through the parents of a location until a BufferSpan is reached, and return the
-        location there.
-
-        So if a location arises from a macro expansion, this will return to the outermost
-        macro invocation.
-        '''
-        while True:
-            span = self.lookup_span(loc)
-            parent_loc = span.macro_parent_loc(loc)
-            if parent_loc == -1:
-                return loc
-            loc = parent_loc
-
     def spelling_coords(self, loc, token_end=False):
         '''Convert a location to a PresumedLocation instance.'''
         span, offset = self.spelling_span_and_offset(loc)
@@ -331,7 +331,7 @@ class Locator:
         '''The length of the token in bytes in the source file.  This incldues, e.g., escaped
         newlines.  The result can be 0, for end-of-source indicator EOF.
         '''
-        lexer = self.pp.lexer_at_loc(loc)
+        lexer = self.lexer_at_loc(loc)
         prior_cursor = lexer.cursor
         lexer.get_token_quietly()
         return lexer.cursor - prior_cursor
@@ -345,15 +345,22 @@ class Locator:
         file, otherwise the outermost macro invocation.  Otherwise use loc directly.
         '''
         if force_outermost_context:
-            loc = self.buffer_span_loc(loc)
-        span, offset = self.spelling_span_and_offset(loc)
+            while True:
+                span = self.lookup_span(loc)
+                parent_range = span.macro_parent_range(loc)
+                if parent_range is None:
+                    break
+                loc = parent_range.start
+            offset = loc - span.start
+        else:
+            span, offset = self.spelling_span_and_offset(loc)
         return span.presumed_location(offset)
 
     def range_coords(self, source_range):
         if isinstance(source_range, SpellingRange):
             # Convert the SpellingRange to a BufferRange
             assert source_range.start < source_range.end
-            lexer = self.pp.lexer_at_loc(source_range.token_loc)
+            lexer = self.lexer_at_loc(source_range.token_loc)
             cursor = lexer.cursor
             lexer.get_token_quietly()
             offsets = [source_range.start, source_range.end]
@@ -366,11 +373,8 @@ class Locator:
             end = self.spelling_coords(source_range.end)
             assert start.buffer is end.buffer
         elif isinstance(source_range, TokenRange):
-            if source_range.start <= location_none:
-                start = end = None
-            else:
-                start = self.spelling_coords(source_range.start)
-                end = self.spelling_coords(source_range.end, True)
+            start = self.spelling_coords(source_range.start)
+            end = self.spelling_coords(source_range.end, True)
         elif source_range is None:
             start = end = None
         else:
@@ -378,46 +382,43 @@ class Locator:
 
         return RangeCoords(start, end)
 
-    def spans_and_locs(self, loc):
+    def spans_and_ranges(self, loc):
         '''Generates the spans of a location stepping up through the locations parents.
 
-        The span of the location is generated first, then the span of its parent location,
+        The span of the location is generated first, then the span of its parent range,
         recursively.
         '''
         result = []
+        range = TokenRange(loc, loc)
         while True:
             span = self.lookup_span(loc)
+            if not span.start <= range.end <= span.end:
+                range.end = span.end
             if isinstance(span, MacroArgumentSpan):
                 macro_span = self.lookup_span(span.parameter_loc)
-                result.append((macro_span, span.parameter_loc))
+                result.append((macro_span, span.macro_parent_range(None)))
                 loc = span.spelling_loc(loc)
+                range = TokenRange(loc, loc)
             else:
-                result.append((span, loc))
-                parent_loc = span.macro_parent_loc(loc)
-                if parent_loc == -1:
+                result.append((span, range))
+                range = span.macro_parent_range(loc)
+                if range is None:
                     return result
-                loc = parent_loc
-
-    def diagnostic_contexts(self, context):
-        # Special diagnostics - those without a source location to highlight, e.g. a
-        # complaint about a command-line option, or a compilation summary - only have a
-        # single location code and no highlight ranges
-        if context.caret_range.start <= location_none:
-            assert not context.source_ranges
-            return [context]
+                loc = range.start
 
+    def diagnostic_contexts(self, main_context):
         def intersections(spans, source_range):
             result = []
             if isinstance(source_range, BufferRange):
-                start_spans_and_locs = self.spans_and_locs(source_range.start)
+                start_spans_and_ranges = self.spans_and_ranges(source_range.start)
                 for span in spans:
-                    for our_span, our_loc in start_spans_and_locs:
+                    for our_span, our_range in start_spans_and_ranges:
                         if span is not our_span:
                             continue
-                        if our_loc == source_range.start:
+                        if our_range.start == source_range.start:
                             item = source_range
                         else:
-                            item = TokenRange(our_loc, our_loc)
+                            item = our_range
                         break
                     else:
                         item = None
@@ -430,33 +431,36 @@ class Locator:
                 # decay into a TokenRange for a single token.
                 token_loc = source_range.token_loc
                 spelling_span, _ = self.spelling_span_and_offset(token_loc)
-                our_spans_and_locs = self.spans_and_locs(token_loc)
+                our_spans_and_ranges = self.spans_and_ranges(token_loc)
                 for span in spans:
-                    for our_span, our_loc in our_spans_and_locs:
+                    for our_span, our_range in our_spans_and_ranges:
                         if span is not our_span:
                             continue
                         if span is spelling_span:
-                            item = SpellingRange(our_loc, source_range.start, source_range.end)
+                            item = SpellingRange(our_range.start, source_range.start,
+                                                 source_range.end)
                         else:
-                            item = TokenRange(our_loc, our_loc)
+                            item = our_range
                         break
                     else:
                         item = None
                     result.append(item)
                 return result
 
-            def token_intersection(span, token_spans_and_locs):
-                for token_span, token_loc in token_spans_and_locs:
+            def token_intersection(span, token_spans_and_ranges, is_start):
+                for token_span, token_range in token_spans_and_ranges:
                     if span is token_span:
-                        return token_loc
+                        loc = token_range.start if is_start else token_range.end
+                        if span.start <= loc <= span.end:
+                            return loc
                 return None
 
             if isinstance(source_range, TokenRange):
-                start_spans_and_locs = self.spans_and_locs(source_range.start)
-                end_spans_and_locs = self.spans_and_locs(source_range.end)
+                start_spans_and_ranges = self.spans_and_ranges(source_range.start)
+                end_spans_and_ranges = self.spans_and_ranges(source_range.end)
                 for span in spans:
-                    start = token_intersection(span, start_spans_and_locs)
-                    end = token_intersection(span, end_spans_and_locs)
+                    start = token_intersection(span, start_spans_and_ranges, True)
+                    end = token_intersection(span, end_spans_and_ranges, False)
                     if start is None:
                         if end is not None:
                             item = TokenRange(span.start, end)
@@ -481,20 +485,19 @@ class Locator:
         # evolution would benefit, the code accepts and handles appropriately any of
         # Buffer Range, SpellingRange, TokenRange or single token locations for the caret
         # range and the highlights.
-        caret_spans_and_locs = self.spans_and_locs(context.caret_range.caret_loc())
-        caret_spans = [span for span, loc in caret_spans_and_locs]
-        caret_ranges = intersections(caret_spans, context.caret_range)
+        caret_spans_and_ranges = self.spans_and_ranges(main_context.caret_range.caret_loc())
+        caret_spans = [span for span, range in caret_spans_and_ranges]
+        caret_ranges = intersections(caret_spans, main_context.caret_range)
         source_ranges_list = [intersections(caret_spans, source_range)
-                              for source_range in context.source_ranges]
+                              for source_range in main_context.source_ranges]
 
-        for n, ((span, caret_loc), caret_range) in enumerate(zip(caret_spans_and_locs,
-                                                                 caret_ranges)):
+        for n, (span, caret_range) in enumerate(zip(caret_spans, caret_ranges)):
             if isinstance(span, BufferSpan):
-                severity = context.severity
-                did, substitutions = context.did, context.substitutions
+                severity = main_context.severity
+                did, substitutions = main_context.did, main_context.substitutions
             else:
                 severity = DiagnosticSeverity.note
-                did, substitutions = span.did_and_substitutions(self.pp, caret_loc)
+                did, substitutions = span.did_and_substitutions(self.pp, caret_range.caret_loc())
 
             source_ranges = [source_ranges_item[n] for source_ranges_item in source_ranges_list]
 
