@@ -10,12 +10,13 @@ from dataclasses import dataclass
 from datetime import datetime
 from enum import IntEnum, auto
 
-from ..diagnostics import DID, Diagnostic
-from .basic import Token, TokenKind, TokenFlags
+from ..core import Token, TokenKind, TokenFlags
+from ..diagnostics import DID, Diagnostic, TokenRange, BufferRange
+
 from .locator import ScratchEntryKind
 
-__all__ = ['Macro', 'MacroFlags', 'ObjectLikeExpansion', 'FunctionLikeExpansion',
-           'expand_builtin_macro', 'predefines']
+__all__ = ['Macro', 'MacroFlags', 'MacroExpansion', 'BuiltinKind', 'expand_builtin_macro',
+           'predefines']
 
 
 class MacroFlags(IntEnum):
@@ -44,6 +45,7 @@ class BuiltinKind(IntEnum):
     TIME = auto()
     FILE = auto()
     LINE = auto()
+    COUNTER = auto()
 
     # _Pragma unary operator
     Pragma = auto()
@@ -104,7 +106,7 @@ class Macro:
         return bool(self.flags & MacroFlags.IS_VARIADIC)
 
     def is_function_like(self):
-        '''Return true if the macro is variadic.'''
+        '''Return true if the macro is function-like.'''
         return bool(self.flags & MacroFlags.IS_FUNCTION_LIKE)
 
     def param_count(self):
@@ -135,308 +137,272 @@ class Macro:
 
         return ''.join(parts(self, pp))
 
-    def collect_arguments(self, pp, name_token):
+    def collect_arguments(self, pp, invocation_loc):
+        '''Return the arguments to the macro.  This is a list of lists of tokens, once list per
+        parameter (including any variable argument).
+        '''
         paren_depth = 0
         get_token = pp.get_token
-        token = Token.create()
         arguments = []
         tokens = []
         param_count = self.param_count()
+        too_many = False
 
-        # Eat the opening parenthesis that was already peeked
-        get_token(token)
-        assert token.kind == TokenKind.PAREN_OPEN
-
-        while True:
-            get_token(token)
-
-            if token.kind == TokenKind.EOF:
-                macro_name = self.macro_name(pp)
-                note = Diagnostic(DID.macro_defined_here, self.name_loc, [macro_name])
-                pp.diag(DID.unterminated_argument_list, name_token.loc, [macro_name, note])
-                arguments = None
-                break
-            if token.kind == TokenKind.PAREN_CLOSE:
-                if paren_depth == 0:
+        # Collect the arguments.  Macro expansion is disabled whilst doing this
+        pp.expand_macros = False
+        with pp.collect_arguments(self):
+            while True:
+                token = get_token()
+                if token.kind == TokenKind.EOF:
+                    macro_name = self.macro_name(pp)
+                    note = Diagnostic(DID.macro_defined_here, self.name_loc, [macro_name])
+                    pp.diag(DID.unterminated_argument_list, invocation_loc, [macro_name, note])
                     break
-                paren_depth -= 1
-            elif paren_depth == 0 and arguments is not None:
-                too_many = False
-                if token.kind == TokenKind.COMMA:
-                    # This completes an argument unless we are in the variable arguments.
-                    if len(arguments) + 1 < param_count:
-                        arguments.append(tokens)
-                        tokens = []
-                        continue
-                    # It is impossible to have too many arguments to a variadic macro; the
-                    # comma just becomes part of the current variable argument.
-                    too_many = not self.is_variadic()
-                else:
-                    too_many = param_count == 0
-                if too_many:
-                    pp.diag(DID.too_many_macro_arguments, token.loc, [self.macro_name(pp)])
-                    arguments = None
-
-            if token.kind == TokenKind.PAREN_OPEN:
-                paren_depth += 1
-            elif token.kind == TokenKind.CONCAT:
-                # Do not produce concat operators from arguments
-                token.kind = TokenKind.OTHER
-
-            # Save the token and continue looking for the ')'.
-            tokens.append(copy(token))
-
-        if arguments is not None:
-            # The ')' completed an argument unless there are no parameters at all.
-            if token.kind == TokenKind.PAREN_CLOSE and param_count:
-                arguments.append(tokens)
-                # Do we have enough arguments?
-                if len(arguments) < param_count:
-                    if len(arguments) < param_count - bool(self.is_variadic()):
-                        pp.diag(DID.too_few_macro_arguments, token.loc, [self.macro_name(pp)])
-                        arguments = None
+                if token.kind == TokenKind.PAREN_CLOSE:
+                    if paren_depth == 0:
+                        break
+                    paren_depth -= 1
+                elif paren_depth == 0 and not too_many:
+                    if token.kind == TokenKind.COMMA:
+                        # This completes an argument unless we are in the variable arguments.
+                        if len(arguments) + 1 < param_count:
+                            arguments.append(tokens)
+                            tokens = []
+                            continue
+                        # It is impossible to have too many arguments to a variadic macro; the
+                        # comma just becomes part of the current variable argument.
+                        too_many = not self.is_variadic()
                     else:
-                        # Add an empty variable argument if none was given
-                        arguments.append([])
-            assert arguments is None or len(arguments) == param_count
+                        too_many = param_count == 0
+                    if too_many:
+                        pp.diag(DID.too_many_macro_arguments, token.loc, [self.macro_name(pp)])
 
-        return arguments
+                if token.kind == TokenKind.PAREN_OPEN:
+                    paren_depth += 1
+                # Save the token and continue looking for the ')'.  Remove leading WS from the
+                # first token.
+                if not tokens:
+                    token.flags &= ~TokenFlags.WS
+                tokens.append(token)
+
+        pp.expand_macros = True
+        # The ')', or EOF, completed an argument unless there are no parameters at all.
+        if param_count:
+            arguments.append(tokens)
+            # Do we have enough arguments?
+            if len(arguments) < param_count - bool(self.is_variadic()):
+                pp.diag(DID.too_few_macro_arguments, token.loc, [self.macro_name(pp)])
+            # Add empty arguments until we have the correct amount
+            while len(arguments) < param_count:
+                arguments.append([])
+
+        assert len(arguments) == param_count
+        return arguments, TokenRange(invocation_loc, token.loc)
 
 
-class SimpleTokenList:
-    '''Common functionaliy for various token lists.'''
+class MacroExpansion:
+    '''A token source that returns the expansion of a macro.'''
 
-    peek_end_kind = TokenKind.PEEK_AGAIN
-
-    def peek_token_kind(self):
-        if self.cursor == len(self.tokens):
-            return self.peek_end_kind
-        return self.tokens[self.cursor].kind
-
-    def concatenate_tokens(self, lhs, concat_loc, rhs):
-        '''Concatenate lhs and rhs; rhs must not be modified.  Return True on success and replace
-        lhs with the result, except lhs must retain its WS flag.  Return False on failure
-        and leave lhs and rhs unmodified.
-
-        concat_loc is the macro location of the ## or %:%: token.
-        '''
-        lhs_spelling = self.pp.token_spelling(lhs)
-        rhs_spelling = self.pp.token_spelling(rhs)
-        spelling = lhs_spelling + rhs_spelling
-
-        if not spelling:
-            # Concatenation of two placemarkers - result is the left placemarker
-            assert lhs.kind == rhs.kind == TokenKind.PLACEMARKER
-            return True
-
-        token, all_consumed = self.pp.lex_from_scratch(spelling, concat_loc,
-                                                       ScratchEntryKind.concatenate)
-        if token.kind == TokenKind.EOF or not all_consumed:
-            self.pp.diag(DID.token_concatenation_failed, concat_loc, [spelling])
-            return False
-        # Do not produce concat operators through concatenation
-        if token.kind == TokenKind.CONCAT:
-            token.kind = TokenKind.OTHER
-        if lhs.flags & TokenFlags.WS:
-            token.flags |= TokenFlags.WS
-        lhs.set_to(token, token.loc)
-        return True
-
-
-class ObjectLikeExpansion(SimpleTokenList):
-    '''A token source that returns tokens from the replacement list of an object-like macro.
-    Token concatenation via the ## operator is handled by get_token().
-    '''
-
-    def __init__(self, pp, macro, parent_token):
-        assert parent_token.kind == TokenKind.IDENTIFIER
-        self.pp = pp
-        self.macro = macro
-        self.parent_flags = parent_token.flags
-        self.tokens = macro.replacement_list
-        self.cursor = 0
-        self.base_loc = pp.locator.macro_replacement_span(macro, parent_token.loc)
-        macro.disable()
-
-    def get_token(self, token):
-        '''Get the next replacement list token and handle token concatenation.'''
-        cursor = self.cursor
-        tokens = self.tokens
-        if cursor == len(tokens):
-            self.macro.enable()
-            self.pp.pop_source_and_get_token(token)
-            return
-
-        token.set_to(tokens[cursor], self.base_loc + cursor)
-        if cursor == 0:
-            # Take whitespace from the invocation token
-            if self.parent_flags & TokenFlags.WS:
-                token.flags |= TokenFlags.WS
-            else:
-                token.flags &= ~TokenFlags.WS
-        cursor += 1
-
-        while cursor != len(tokens) and tokens[cursor].kind == TokenKind.CONCAT:
-            assert cursor + 1 < len(tokens)
-            if self.concatenate_tokens(token, self.base_loc + cursor, tokens[cursor + 1]):
-                cursor += 2
-            else:
-                cursor += 1
-        self.cursor = cursor
-
-
-class FunctionLikeExpansion(SimpleTokenList):
-    '''A token source that returns tokens from the replacement list of an function-like macro.
-    Token concatenation via the ## operator, and stringizing via the # operator, are
-    handled by get_token().
-    '''
-
-    def __init__(self, pp, macro, parent_token, arguments):
-        assert parent_token.kind == TokenKind.IDENTIFIER
+    def __init__(self, pp, macro, invocation_loc, token_ws_flag):
         self.pp = pp
         self.macro = macro
         self.cursor = 0
-        self.trailing_ws = False
         tokens = macro.replacement_list
-        base_loc = pp.locator.macro_replacement_span(macro, parent_token.loc)
-        self.tokens = self.replace_arguments(tokens, arguments, base_loc, 0, len(tokens),
-                                             parent_token.flags)
+        if macro.is_function_like():
+            arguments, invocation_range = macro.collect_arguments(pp, invocation_loc)
+        else:
+            invocation_range = TokenRange(invocation_loc, invocation_loc)
+            arguments = None
+        base_loc = pp.locator.macro_replacement_span(macro, invocation_range)
+        # Small optimization: having checked the arguments, expand a function-like macro
+        # taking no arguments as if it were object-like
+        if arguments:
+            self.tokens = self.replace_arguments(tokens, arguments, base_loc, 0, len(tokens),
+                                                 token_ws_flag)
+        else:
+            self.tokens = self.objlike_replacement_tokens(tokens, base_loc, token_ws_flag)
         macro.disable()
 
-    def replace_arguments(self, tokens, arguments, base_loc, cursor, limit, parent_flags):
-        def check_argument(param):
-            assert param.kind == TokenKind.MACRO_PARAM
-            if param.extra < 0:
-                assert self.macro.is_variadic()
-                # Pre-expand the variable argument to see if it's empty.
-                va_tokens = self.expand_argument(arguments[-1])
-                # FIXME: remove placemarkers
-                if va_tokens:
-                    token_count = -param.extra   # Includes the parentheses
-                    va_opt_tokens = self.replace_arguments(
-                        tokens, arguments, base_loc, cursor + 2, cursor + token_count,
-                        param.flags)
-                    return va_opt_tokens, token_count
-                else:
-                    return [self.placemarker_token()], -param.extra
-            else:
-                assert 0 <= param.extra < len(arguments)
-                return arguments[param.extra], 0
-
-        result = []
-        loop_count = -1
-        ws = parent_flags & TokenFlags.WS
-        while cursor < limit:
-            loop_count += 1
-            token = tokens[cursor]
-            new_token_loc = base_loc + cursor
-            if loop_count and token.flags & TokenFlags.WS:
-                ws = True
-
-            if token.kind == TokenKind.STRINGIZE:
-                cursor += 1
-                argument_tokens, va_opt_count = check_argument(tokens[cursor])
-                string = self.stringize_argument(argument_tokens, new_token_loc)
-                if ws:
-                    string.flags |= TokenFlags.WS
-                    ws = False
-                result.append(string)
-                cursor += 1 + va_opt_count
-                continue
-
-            if token.kind == TokenKind.MACRO_PARAM:
-                argument_tokens, token_count = check_argument(token)
-                cursor += token_count + 1
-                lhs_concat = result and result[-1].kind == TokenKind.CONCAT
-                rhs_concat = (cursor != limit and tokens[cursor].kind == TokenKind.CONCAT)
-                if lhs_concat or rhs_concat or token_count:
-                    if argument_tokens:
-                        argument_tokens = [copy(token) for token in argument_tokens]
-                    # Replace empty argument with a placemarker
-                    if not argument_tokens:
-                        argument_tokens = [self.placemarker_token()]
-                else:
-                    argument_tokens = self.expand_argument(argument_tokens)
-
-                # Set WS flag on the first token (if there is one)
-                if argument_tokens:
-                    if ws:
-                        argument_tokens[0].flags |= TokenFlags.WS
-                        ws = False
-                    elif loop_count == 0:
-                        argument_tokens[0].flags &= ~TokenFlags.WS
-                # Give the tokens their macro-expansion locations
-                locations = [token.loc for token in argument_tokens]
-                first_loc = self.pp.locator.macro_argument_span(new_token_loc, locations)
-                for loc, token in enumerate(argument_tokens, start=first_loc):
-                    token.loc = loc
-
-                result.extend(argument_tokens)
-                continue
-
-            token = copy(token)
-            if ws:
-                token.flags |= TokenFlags.WS
-                ws = False
-            elif loop_count == 0:
-                token.flags &= ~TokenFlags.WS
-            token.loc = new_token_loc
-            result.append(token)
-            cursor += 1
-
-        result, result_ws = self.perform_concatenations(result)
-        self.trailing_ws = ws or result_ws
-        return result
-
-    def perform_concatenations(self, tokens):
-        result = []
-        cursor = 0
-        limit = len(tokens)
-        ws = False
-        while cursor < limit:
-            token = tokens[cursor]
-            if token.kind == TokenKind.CONCAT:
-                assert cursor + 1 < len(tokens)
-                if self.concatenate_tokens(result[-1], token.loc, tokens[cursor + 1]):
-                    cursor += 2
-                else:
-                    cursor += 1
-            else:
-                result.append(token)
-                cursor += 1
-            if ws:
-                result[-1].flags |= TokenFlags.WS
-                ws = False
-
-        return result, ws
-
-    def get_token(self, token):
+    def get_token(self):
         '''Handle argument replacement, stringizing and token concatenation.'''
         cursor = self.cursor
         tokens = self.tokens
         if cursor == len(tokens):
             self.macro.enable()
-            self.pp.pop_source_and_get_token(token)
-            if self.trailing_ws:
-                token.flags |= TokenFlags.WS
-            return
+            token = self.pp.pop_source_and_get_token()
+            token.flags |= self.trailing_ws
+            return token
 
-        token.set_to(tokens[cursor], tokens[cursor].loc)
+        token = copy(tokens[cursor])
         self.cursor = cursor + 1
+        return token
 
-        # Don't let placemarker tokens leak out
-        if token.kind == TokenKind.PLACEMARKER:
-            ws = bool(token.flags & TokenFlags.WS)
-            self.get_token(token)
-            if ws:
-                token.flags |= TokenFlags.WS
+    def peek_token_kind(self):
+        if self.cursor == len(self.tokens):
+            return TokenKind.PEEK_AGAIN
+        return self.tokens[self.cursor].kind
+
+    def concatenate_tokens(self, tokens, concat_loc, rhs):
+        '''Concatenate tokens[-1] and rhs.  On success replace tokens[-1] with the result and the
+        WS flag of tokens[-1], otherwise append rhs to tokens.
+        concat_loc is the macro location of the ## or %:%: token.
+        '''
+        lhs_spelling = self.pp.token_spelling(tokens[-1])
+        rhs_spelling = self.pp.token_spelling(rhs)
+        spelling = lhs_spelling + rhs_spelling
+
+        if spelling:
+            parent_range = TokenRange(tokens[-1].loc, rhs.loc)
+            token, all_consumed = self.pp.lex_from_scratch(spelling, parent_range,
+                                                           ScratchEntryKind.concatenate)
+            if token.kind == TokenKind.EOF or not all_consumed:
+                diag_loc = BufferRange(token.loc, token.loc + len(spelling))
+                self.pp.diag(DID.token_concatenation_failed, diag_loc, [spelling])
+                # Make a copy of RHS and give it the correct location
+                rhs = copy(rhs)
+                rhs.loc = concat_loc + 1
+                tokens.append(rhs)
+            else:
+                if tokens[-1].flags & TokenFlags.WS:
+                    token.flags |= TokenFlags.WS
+                tokens[-1] = token
+        else:  # Concatenation of two placemarkers - result is the left placemarker
+            assert tokens[-1].kind == rhs.kind == TokenKind.PLACEMARKER
+
+    def objlike_replacement_tokens(self, tokens, base_loc, ws):
+        '''Copy the replacement list tokens, giving them their new locations.  Whilst doing so,
+        perform any concatenations present.  Return a token list ready for stepping
+        through with get_token().
+        '''
+        # Copy the replacement list tokens giving them their new in-expansion locations.
+        result = []
+        cursor, limit = 0, len(tokens)
+        while cursor < limit:
+            token = copy(tokens[cursor])
+            token.loc = base_loc + cursor
+            if token.kind == TokenKind.CONCAT:
+                assert cursor + 1 < limit
+                self.concatenate_tokens(result, token.loc, tokens[cursor + 1])
+                cursor += 2
+            else:
+                result.append(token)
+                cursor += 1
+
+        # The first token acquires the WS flag of the invocation token.  If there is no
+        # token, record any trailing whitepsace.
+        if result:
+            result[0].flags &= ~TokenFlags.WS
+            result[0].flags |= ws
+            ws = 0
+        self.trailing_ws = ws
+        return result
+
+    def replace_arguments(self, tokens, arguments, base_loc, cursor, limit, ws):
+        result = []
+        first = cursor
+        while cursor < limit:
+            token = tokens[cursor]
+            new_token_loc = base_loc + cursor
+            if cursor != first:
+                ws |= token.flags & TokenFlags.WS
+
+            if token.kind != TokenKind.MACRO_PARAM:
+                token = copy(token)
+                token.loc = new_token_loc
+                # Apply whitepace appropriately
+                if cursor == first:
+                    token.flags &= ~TokenFlags.WS
+                token.flags |= ws
+                ws = 0
+                result.append(token)
+                cursor += 1
+                continue
+
+            # Determine the argument
+            if token.extra >= 0:
+                argument_tokens = arguments[token.extra]
+                va_opt_count = 0
+            else:  # __VA_OPT__
+                assert self.macro.is_variadic()
+                # Fully expand the variable argument only to see if it's empty
+                va_opt_count = -token.extra   # Includes the parentheses
+                va_tokens = self.expand_argument(arguments[-1])
+                if va_tokens:
+                    argument_tokens = self.replace_arguments(
+                        tokens, arguments, base_loc, cursor + 2, cursor + va_opt_count, 0)
+                else:
+                    argument_tokens = [self.placemarker_token()]
+            cursor += 1 + va_opt_count
+
+            # Stringify the argument?
+            if result and result[-1].kind == TokenKind.STRINGIZE:
+                stringize_range = TokenRange(result[-1].loc, new_token_loc)
+                token = self.stringize_argument(argument_tokens, stringize_range)
+                assert not (token.flags & TokenFlags.WS)
+                token.flags |= result[-1].flags & TokenFlags.WS
+                result[-1] = token
+                ws = 0   # Don't carry the whitespace of the parameter
+            else:
+                lhs_concat = result and result[-1].kind == TokenKind.CONCAT
+                rhs_concat = (cursor != limit and tokens[cursor].kind == TokenKind.CONCAT)
+                if lhs_concat or rhs_concat or va_opt_count:
+                    if argument_tokens:
+                        argument_tokens = [copy(token) for token in argument_tokens]
+                    else:
+                        # Empty argument in a concatenation becomes a placemarker
+                        argument_tokens = [self.placemarker_token()]
+                else:
+                    argument_tokens = self.expand_argument(argument_tokens)
+                # Set WS flag on the first token if there is one
+                if argument_tokens:
+                    argument_tokens[0].flags |= ws
+                    ws = 0
+                # Give the argument tokens their special locations
+                locations = [token.loc for token in argument_tokens]
+                first_loc = self.pp.locator.macro_argument_span(new_token_loc, locations)
+                for loc, token in enumerate(argument_tokens, start=first_loc):
+                    token.loc = loc
+                result.extend(argument_tokens)
+
+        result, result_ws = self.perform_concatenations(result, first == 0)
+        if first == 0:
+            self.trailing_ws = ws | result_ws
+        return result
+
+    def perform_concatenations(self, tokens, remove_placemarkers):
+        '''Perform concatenations for a function-like macro expansion before the rescan
+        (get_token) step.  On leaving this function no placemarker tokens should remain.
+        '''
+        result = []
+        cursor = 0
+        limit = len(tokens)
+        ws = 0
+        while cursor < limit:
+            token = tokens[cursor]
+            if token.kind == TokenKind.CONCAT:
+                assert cursor + 1 < limit
+                self.concatenate_tokens(result, token.loc, tokens[cursor + 1])
+                cursor += 2
+            else:
+                if remove_placemarkers and result and result[-1].kind == TokenKind.PLACEMARKER:
+                    ws |= result[-1].flags & TokenFlags.WS
+                    result.pop()
+                result.append(token)
+                cursor += 1
+            result[-1].flags |= ws
+            ws = 0
+
+        if remove_placemarkers and result and result[-1].kind == TokenKind.PLACEMARKER:
+            ws |= result[-1].flags & TokenFlags.WS
+            result.pop()
+
+        return result, ws
 
     def expand_argument(self, argument_tokens):
         def collect_expanded_tokens(get_token):
             result = []
             while True:
-                token = Token.create()
-                get_token(token)
+                token = get_token()
                 if token.kind == TokenKind.EOF:
                     return result
                 result.append(token)
@@ -449,7 +415,7 @@ class FunctionLikeExpansion(SimpleTokenList):
     def placemarker_token(self):
         return Token(TokenKind.PLACEMARKER, 0, 0, None)
 
-    def stringize_argument(self, argument_tokens, stringize_loc):
+    def stringize_argument(self, argument_tokens, stringize_range):
         '''Stringize the argument tokens.  Return a new token.  stringize_loc is the macro
         location of the # operator.
 
@@ -461,15 +427,15 @@ class FunctionLikeExpansion(SimpleTokenList):
 
         # Build up the spelling of the stringized argument tokens.
         spelling.append(34)    # '"'
-        ws = False
+        ws = 0
         for n, token in enumerate(argument_tokens):
-            if n and token.flags & TokenFlags.WS:
-                ws = True
+            if n:
+                ws |= token.flags & TokenFlags.WS
             if token.kind == TokenKind.PLACEMARKER:
                 continue   # Placemarkers can only contribute to whitespace and not accumulate it
             if ws:
                 spelling.append(32)
-                ws = False
+                ws = 0
             token_spelling = self.pp.token_spelling(token)
             if token.kind == TokenKind.STRING_LITERAL or token.kind == TokenKind.CHARACTER_LITERAL:
                 for c in token_spelling:
@@ -483,38 +449,33 @@ class FunctionLikeExpansion(SimpleTokenList):
             else:
                 spelling.extend(token_spelling)
         spelling.append(34)    # '"'
-        token, all_consumed = self.pp.lex_from_scratch(spelling, stringize_loc,
+        token, all_consumed = self.pp.lex_from_scratch(spelling, stringize_range,
                                                        ScratchEntryKind.stringize)
-        assert all_consumed
-
-        if token.kind == TokenKind.ERROR:
-            self.pp.diag(DID.stringize_failed, stringize_loc)
-            # Replace with an empty string literal.  FIXME: imitate GCC and Clang?
-            token = Token(TokenKind.STRING_LITERAL, 0, stringize_loc, (b'""', None))
-
-        assert token.kind == TokenKind.STRING_LITERAL
-        return token
+        return token   # kind is either STRING_LITERAL or UNTERMINATED
 
 
-class UnexpandedArgument(SimpleTokenList):
-
-    peek_end_kind = TokenKind.EOF
+class UnexpandedArgument:
 
     def __init__(self, tokens):
         self.tokens = tokens
         self.cursor = 0
 
-    def get_token(self, token):
+    def peek_token_kind(self):
+        if self.cursor == len(self.tokens):
+            return TokenKind.EOF
+        return self.tokens[self.cursor].kind
+
+    def get_token(self):
         cursor = self.cursor
         tokens = self.tokens
         if cursor == len(tokens):
             # Terminate the collect_expanded_tokens() loop.
-            token.kind = TokenKind.EOF
-        else:
-            # Don't care about the whitespace of the first token - it is set by
-            # FunctionLikeExpansion.get_token().
-            token.set_to(tokens[cursor], tokens[cursor].loc)
-            self.cursor = cursor + 1
+            return Token(TokenKind.EOF, 0, 0, None)
+
+        # Don't care about the whitespace of the first token - it is set by
+        # FunctionLikeExpansion.get_token().
+        self.cursor = cursor + 1
+        return tokens[cursor]
 
 
 def expand_builtin_macro(pp, token):
@@ -533,9 +494,9 @@ def expand_builtin_macro(pp, token):
         if pp.time_str is None:
             epoch = pp.source_date_epoch
             if epoch is None:
-                epoch = datetime.today()
+                epoch = datetime.now(tz=pp.tz)
             else:
-                epoch = datetime.fromtimestamp(epoch)
+                epoch = datetime.fromtimestamp(epoch, tz=pp.tz)
             months = 'Jan Feb Mar Apr May Jun Jul Aug Sep Oct Nov Dec'.split()
             pp.time_str = f'"{epoch.hour:02d}:{epoch.minute:02d}:{epoch.second:02d}"'
             pp.date_str = f'"{months[epoch.month - 1]} {epoch.day:2d} {epoch.year:4d}"'
@@ -544,21 +505,25 @@ def expand_builtin_macro(pp, token):
             spelling = pp.time_str
         else:
             spelling = pp.date_str
+    elif kind == BuiltinKind.COUNTER:
+        spelling = str(pp.counter)
+        pp.counter += 1
     else:
-        assert False
+        pp.ice(token.loc, f'unhandled builtin kind {kind}')
+        return token
 
-    lex_token_from_builtin_spelling(pp, token, spelling)
+    return lex_token_from_builtin_spelling(pp, token, spelling, TokenRange(token.loc, token.loc))
 
 
-def lex_token_from_builtin_spelling(pp, token, spelling):
+def lex_token_from_builtin_spelling(pp, token, spelling, parent_range):
     ws = token.flags & TokenFlags.WS
-    btoken, all_consumed = pp.lex_from_scratch(spelling.encode(), token.loc,
-                                               ScratchEntryKind.builtin)
+    assert spelling
+    token, all_consumed = pp.lex_from_scratch(spelling.encode(), parent_range,
+                                              ScratchEntryKind.builtin)
     assert all_consumed
-    assert not (btoken.flags & TokenFlags.WS)
-    if ws:
-        btoken.flags |= TokenFlags.WS
-    token.set_to(btoken, btoken.loc)
+    assert not (token.flags & TokenFlags.WS)
+    token.flags |= ws
+    return token
 
 
 def predefines(pp):
@@ -617,9 +582,9 @@ def standard_c_predefines(pp):
 
 
 def standard_cxx_predefines(pp):
-    # We do not define __STDC__ or __STDC_VERSION__ or __STDC_ISO_10646__ when compiling
-    # C++.
     yield '__cplusplus', '202302L'
+    yield '__STDC__', '1'
+    # We do not define __STDC_VERSION__ or __STDC_ISO_10646__ when compiling C++.
 
     # The values for C++23
     yield '__cpp_aggregate_bases', '201603L'
