@@ -9,32 +9,34 @@ from bisect import bisect_left
 from dataclasses import dataclass
 from itertools import accumulate
 
-from ..core import Buffer, PresumedLocation
+from ..core import Buffer, PresumedLocation, host
 from ..unicode import (
     utf8_cp, is_printable, terminal_charwidth, codepoint_to_hex,
 )
-from .diagnostic import Diagnostic, DiagnosticEngine
+from .diagnostic import Diagnostic, DiagnosticConsumer
 
 
 __all__ = ['UnicodeTerminal']
 
 
-class UnicodeTerminal(DiagnosticEngine):
+class UnicodeTerminal(DiagnosticConsumer):
     '''Write formatted diagnostics to stderr, in a way that they should be suitable for
     display on a Unicode-enabled terminal.
     '''
+    elaborate = True
 
-    def __init__(self, pp, *, translations=None):
+    def __init__(self):
         '''Diagnostics are written to file.  Colour formatting information, whether colours are
         enabled, and the tabstop are taken from env.  Source file tabs are space-expanded
         to the tabstop.  Diagnostics are adjusted for the terminal width, which we attempt
         to determine from file.
         '''
-        super().__init__(pp, translations=translations)
+        super().__init__()
         self.nested_indent = 4
         self.sgr_codes = {}
         self.tabstop = 8
-        self.terminal_width = 120
+        # Take from stderr terminal (if it is a terminal)
+        self.terminal_width = 0
 
     def set_sgr_code_assignments(self, colour_string):
         '''Parse an SGR assignments string.'''
@@ -55,34 +57,149 @@ class UnicodeTerminal(DiagnosticEngine):
 
     def emit(self, diagnostic: Diagnostic):
         '''Called when the preprocessor emits a diagnostic.'''
-        # Elaborate it, and emit it recursively.
-        self.emit_recursive(self.elaborate(diagnostic), 0)
+        # Configure the terminal width if not already set
+        if self.terminal_width <= 0:
+            if host.is_a_tty(self.stderr):
+                self.terminal_width = host.terminal_width(self.stderr)
+            else:
+                self.terminal_width = 120
+        self.emit_recursive(diagnostic, 0)
 
-    def emit_recursive(self, elaborated_diagnostic, indent):
+    def emit_recursive(self, diagnostic, indent):
         '''Emit the top-level diagnostic at the given indentation level.  Then emit nested
         diagnostics at an increased indentation level.
         '''
         orig_indent = indent
-        for n, message_context in enumerate(elaborated_diagnostic.message_contexts):
+        for n, message_context in enumerate(diagnostic.message_contexts):
             if n == 1:
                 indent += self.nested_indent
-            for line in self.diagnostic_lines(message_context):
-                print(f'{" " * indent}{line}', file=self.pp.stderr)
-        for nested in elaborated_diagnostic.nested_diagnostics:
+            room = self.terminal_width - 1 - indent
+            for line in self.diagnostic_lines(message_context, room):
+                print(f'{" " * indent}{line}', file=self.stderr)
+        for nested in diagnostic.nested_diagnostics:
             self.emit_recursive(nested, orig_indent + self.nested_indent)
 
-    def diagnostic_lines(self, context):
+    def diagnostic_lines(self, context, room):
         '''Generate all the lines to display for the diagnostic conext - one for the message, and
         perhaps several source lines and highlights.
         '''
         # The main diagnostic message, perhaps prefixed with file and line number, and
-        # diagnostic severity, that let's the user know what the complaint is.
-        yield ''.join(self.enhance_text(*part) for part in context.message_parts)
+        # diagnostic severity, that informs the user know what the complaint is.
+        yield from self.word_wrapped_lines(context.message_parts, room)
 
         if context.caret_highlight.start is not None:
-            yield from self.show_highlighted_source(context)
+            yield from self.show_highlighted_source(context, room)
 
-    def show_highlighted_source(self, context):
+    def word_wrapped_lines(self, parts, room):
+        def accept_text(accepted_text, new_text, new_width, mid_line):
+            if not mid_line:
+                # Strip leading space from what will begin the next line
+                count = 0
+                while count < len(new_text) and new_text[count] == ' ':
+                    count += 1
+                new_width -= count
+                new_text = new_text[count:]
+            accepted_text += new_text
+            return accepted_text, new_width
+
+        def lines(parts, room, indent):
+            # Return a sequence of parts to go on a line.  Subsequent lines are indented
+            # by indent, and the first part of those lines consists of the spaces.
+            room = max(room, indent + 2)
+            result = []
+            line = []
+            accepted_parts = False
+            accepted_width = 0
+            for text, hint in parts:
+                # Strip the text at the start of a line
+                if not accepted_parts:
+                    text = text.lstrip()
+                if not text:
+                    # Substitutions can also leave empty parts
+                    continue
+                char_widths = bytes(terminal_charwidth(ord(c)) for c in text)
+                maybe_width = sum(char_widths)
+                if accepted_width + maybe_width <= room:
+                    line.append((text, hint))
+                    accepted_width += maybe_width
+                    accepted_parts = True
+                    continue
+                # If it can't be broken up, flush the current line if we have something and
+                # start the next line with this part
+                if hint != 'message':
+                    if accepted_parts:
+                        result.append(line)
+                        line = [(' ' * indent, 'message'), (text, hint)]
+                        accepted_width = indent + maybe_width
+                    else:
+                        line.append((text, hint))
+                        accepted_width += maybe_width
+                        accepted_parts = True
+                    continue
+                # Break the message up at spaces (or at CJK characters) over as many lines
+                # as needed.  Last break point is the first text character which can be
+                # pushed to the next line if a break is needed
+                accepted_text = ''
+                maybe_text = ''
+                maybe_width = 0
+                prev_char_width = 0
+                for n, (c, char_width) in enumerate(zip(text, char_widths)):
+                    # If this is a break position (before a CJK character or space, or
+                    # after a CJK character), add maybe_text to accepted_text
+                    if c == ' ' or char_width == 2 or prev_char_width == 2:
+                        accepted_text, maybe_width = accept_text(accepted_text, maybe_text,
+                                                                 maybe_width, accepted_parts)
+                        accepted_width += maybe_width
+                        accepted_parts = accepted_parts or maybe_width
+                        maybe_text = ''
+                        maybe_width = 0
+                    prev_char_width = char_width
+                    # Add this character to the text whose line we're not yet sure of
+                    maybe_text += c
+                    maybe_width += char_width
+                    # If it fits or we've not yet accepted something, continue
+                    if accepted_width + maybe_width <= room or not accepted_parts:
+                        continue
+                    # It doesn't fit and we have a non-empty line.  Flush the line and
+                    # begin the next line with the indent.  accepted_text can be empty
+                    # because of accepted parts before trying to split this 'message'
+                    if accepted_text:
+                        line.append((accepted_text, 'message'))
+                    result.append(line)
+                    line = [(' ' * indent, 'message')]
+                    accepted_text = ''
+                    accepted_width = indent
+                    accepted_parts = False
+                accepted_text, maybe_width = accept_text(accepted_text, maybe_text,
+                                                         maybe_width, accepted_parts)
+                accepted_width += maybe_width
+                accepted_parts = accepted_parts or maybe_width
+                if accepted_text:
+                    line.append((accepted_text, 'message'))
+            if line:
+                result.append(line)
+            return result
+
+        def enhance_line(parts):
+            '''Combine hints to enhance text in groups.'''
+            hint = None
+            combined_text = ''
+            line = ''
+            for text, text_hint in parts:
+                if hint is None or text_hint == hint:
+                    combined_text += text
+                    hint = text_hint
+                else:
+                    line += self.enhance_text(combined_text, hint)
+                    combined_text = text
+                    hint = text_hint
+            line += self.enhance_text(combined_text, hint)
+            return line
+
+        for line in lines(parts, room, 10):
+            yield enhance_line(line)
+
+    def show_highlighted_source(self, context, room):
         '''The first highlight lies in a single buffer and is the "centre" of the diagnostic where
         the caret is shown.  This function generates a sequence of text lines:
 
@@ -96,21 +213,21 @@ class UnicodeTerminal(DiagnosticEngine):
         Other highlights appear only to the extent the intersect the first highlight's
         source lines.
         '''
-        def line_margins(line_number, number_width):
-            '''Return the text margin to show at the beginning of the line.'''
-            margin = f'{line_number:{number_width}d}'
-            return margin + ' | ', ' ' * len(margin) + ' | '
-
         start, end = context.caret_highlight.start, context.caret_highlight.end
         lines = self.source_lines(start, end)
+
+        # Margin book-keeping details
         max_line_number = start.presumed_line_number + len(lines) - 1
-        number_width = max(5, len(str(max_line_number)))
+        line_number_width = max(5, len(str(max_line_number)))
+        margin_suffix = ' | '
+        margin_width = len(margin_suffix) + line_number_width
+        room -= margin_width
+
         for line_number, line in enumerate(lines, start=start.presumed_line_number):
-            margins = line_margins(line_number, number_width)
-            room = self.terminal_width - 1 - len(margins[0])
+            margins = f'{line_number:{line_number_width}d}', ' ' * line_number_width
             texts = line.source_and_highlight_lines(context, room, self.enhance_text)
             for margin, text in zip(margins, texts):
-                yield margin + text
+                yield margin + margin_suffix + text
 
     def source_lines(self, start: PresumedLocation, end: PresumedLocation):
         '''Return a list of SourceLine objects, one for each line between start and end inclusive.
@@ -328,6 +445,8 @@ class SourceLine:
             # Drop highlights that don't appear on this line
             if start == -1:
                 continue
+            start = max(0, min(start - removed_chars, room))
+            end = max(0, min(end - removed_chars, room))
             # Special handling of caret range - the first character (including if wide) gets
             # the caret, the rest of the range gets the twiddles
             if n == 0:
@@ -361,12 +480,12 @@ class SourceLine:
 
         def source_line_parts(line):
             cursor = 0
-            for r in self.replacements:
-                ncursor = sum(self.out_widths[:r])
-                yield self.text[cursor: ncursor]
-                cursor = ncursor + self.out_widths[r]
-                replacement = self.text[ncursor:cursor]
+            for r in line.replacements:
+                ncursor = sum(line.out_widths[:r])
+                yield line.text[cursor: ncursor]
+                cursor = ncursor + line.out_widths[r]
+                replacement = line.text[ncursor:cursor]
                 yield enhance_text(replacement, 'unprintable')
-            yield self.text[cursor:]
+            yield line.text[cursor:]
 
         return ''.join(source_line_parts(line)), ''.join(highlight_parts(char_ranges))
